from munch import Munch
from functools import partial

from infi.shell import commandgroup, command
from infi.shell.utils import strlist
from infi.shell.exceptions import BadArguments

from .types import Name, Host, Cluster, Volume, MasterVolume, VolumeSnap, VolumeClone, Lun
from .options import QueryOptions
from .host import HostCommandsBase
from . import render_collection, roles

QUERY_FIELDS = strlist("name hosts luns created_at")
DETAILED_QUERY_FIELDS = strlist("name hosts luns created_at updated_at")
SORTABLE_FIELDS = strlist("name created_at updated_at")

HOST_QUERY_FIELDS = strlist("name host")
MAP_QUERY_FIELDS = strlist("cluster volume lun")

@commandgroup(prefix="cluster", provides=(Cluster,))
class ClusterCommands(HostCommandsBase):
    ENTITY_TYPE = Cluster

    @command(
        description="Create a new cluster",
        types=(Name,),
        )
    @roles.INFINIDAT.ADMIN.POOLADMIN
    def create(self, _api, name):
        return self._create(_api, name=name)

    @command(
        description="Rename cluster",
        types=(Cluster, Name),
        substitute=dict(cluster="name"),
        )
    @roles.INFINIDAT.ADMIN.POOLADMIN
    def rename(self, _api, cluster, new_name):
        return self._rename(_api, cluster, new_name)

    @command(
        description="Delete cluster",
        types=(Cluster,),
        substitute=dict(cluster="name"),
        )
    @roles.INFINIDAT.ADMIN.POOLADMIN
    def delete(self, _api, cluster):
        return self._delete(_api, cluster)

    @command(
        description="Add host to cluster",
        types=(Cluster, Host.List),
        substitute=dict(
            cluster="name",
            hosts="host",
            ),
        )
    @roles.INFINIDAT.ADMIN.POOLADMIN
    def add_host(self, _api, cluster, hosts):
        with _api.aggregation() as aggregator:
            for host in hosts:
                aggregator.next_task(host.name, "Added")
                self._post(_api, cluster, "hosts", data=dict(id=host.id))
        if len(hosts) == 1:
            return 'Host "{}" added to cluster "{}"'.format(hosts[0].name, cluster.name)
        return '{} hosts added to cluster "{}"'.format(len(hosts), cluster.name)

    @command(
        description="Remove host from cluster",
        types=(Cluster, Host.List),
        substitute=dict(
            cluster="name",
            hosts="host",
            ),
        )
    @roles.INFINIDAT.ADMIN.POOLADMIN
    def remove_host(self, _api, cluster, hosts):
        with _api.aggregation() as aggregator:
            for host in hosts:
                aggregator.next_task(host.name, "Removed")
                self._del(_api, cluster, "hosts", host.id)
        if len(hosts) == 1:
            return 'Host "{}" removed from cluster "{}"'.format(host.name, cluster.name)
        return '{} hosts removed from cluster "{}"'.format(len(hosts), cluster.name)

    @command(
        description="Map {} to a cluster".format(Volume.DESCRIPTION_SUBJECT),
        types=(Cluster, MasterVolume.List, VolumeSnap.List, VolumeClone.List, Lun),
        substitute=dict(cluster="name"),
        )
    @roles.INFINIDAT.ADMIN.POOLADMIN
    def map(self, _api, cluster, vol=None, snap=None, clone=None, lun=None):
        return self._map(_api, cluster, lun, vol, snap, clone)

    @command(
        description="Unmap {} from cluster".format(Volume.DESCRIPTION_SUBJECT),
        types=(Cluster, MasterVolume.List, VolumeSnap.List, VolumeClone.List, Lun),
        substitute=dict(cluster="name"),
        )
    @roles.INFINIDAT.ADMIN.POOLADMIN
    def unmap(self, _api, cluster, vol=None, snap=None, clone=None, lun=None):
        return self._unmap(_api, cluster, lun, vol, snap, clone)

    @command(
        description="List existing clusters",
        types=(Cluster.List,),
        substitute=dict(cluster="name"),
        options=QueryOptions(fields=DETAILED_QUERY_FIELDS, sortable=SORTABLE_FIELDS, detailed=True),
        )
    @render_collection(
        entity=Cluster,
        fields=QUERY_FIELDS,
        alignment=dict(
            hosts="right",
            ),
        detailed_individually=True,
        detailed_fields=DETAILED_QUERY_FIELDS,
        )
    def query(self, _api, cluster=None):
        return self._query(_api, cluster, _hook=self._query_hook)

    def _query_hook(self, clusters):
        for cluster in clusters:
            cluster.hosts = len(cluster.hosts)
            cluster.luns = len(cluster.luns)
        return clusters

    @command(
        description="List existing clusters with member hosts",
        types=(Cluster.List, Host.List),
        substitute=dict(cluster="name"),
        options=QueryOptions(fields=HOST_QUERY_FIELDS, sortable=HOST_QUERY_FIELDS, detailed=False),
        )
    @render_collection(
        entity=Cluster,
        )
    def host_query(self, _api, cluster=None, host=None):
        if cluster and host:
            raise BadArguments("Either cluster, host, or neither may be specified")
        host_names = set(h.name for h in host) if host else None
        return self._query(_api,
            cluster,
            _hook=partial(self._host_query_hook, host_names=host_names),
            _sort_on_client_side=True,
            )

    def _host_query_hook(self, clusters, host_names):
        return [Munch(
            name=cluster.name,
            host=host.name
            )
            for cluster in clusters
            for host in sorted(cluster.hosts, key=lambda h: h.name)
            if not host_names or host.name in host_names]

    @command(
        description="List existing cluster mappings",
        types=(Cluster,),
        substitute=dict(cluster="name"),
        options=QueryOptions(fields=MAP_QUERY_FIELDS),
        )
    @render_collection(
        entity=Cluster,
        fields=MAP_QUERY_FIELDS,
        substitute=dict(
            lun="lun_id",
            ),
        )
    def map_query(self, _api, cluster=None):
        return self._map_query(_api, cluster)
