from collections import OrderedDict

from infi.shell.commands import commandgroup, command, render
from infi.shell.types import (
    IPAddress, Hostname, YesNo, Str
    )

from ..types import Name, EthPort
from .. import roles
from . import ConfigCommandsBase

@commandgroup(parent=ConfigCommandsBase, prefix="system")
class SystemConfigCommands(ConfigCommandsBase):

    def _format_ips_string(self, s):
        return (s or "").replace(",", ", ")

    @command(
        description="Display general system configuration",
        )
    @render(
        individually=True,
        )
    def query(self, _api):
        config = self._get_config(_api)
        result = OrderedDict((
            ("name", _api.get("system/name")),
            ("management_ip_address", "{} / {}".format(config.ip_config.ip, config.ip_config.netmask)),
            ("default_gateway", config.ip_config.gateway),
            ("ssl_Redirection", "{}".format("enabled" if config.mgmt['network.force_ssl_all'] else "disabled")),
            ))
        node_ips = []
        for port in EthPort.iter_all(_api):
            if port.role and port.role.lower() == "management":
                node_ips.append((port.node_index, port.ip_v4_addr))
        result.update(("node_{}_ip_address".format(node_index), ip) for node_index, ip in sorted(node_ips))
        result.update((
            ("dns_servers", self._format_ips_string(config.mgmt["environment.dns_servers"])),
            ("ntp_servers", self._format_ips_string(config.mgmt["environment.ntp_servers"])),
            ))
        return result;

    @command(
        description="Set DNS server IPs",
        types=(IPAddress, IPAddress),
        )
    @roles.INFINIDAT.ADMIN
    def set_dns(self, _api, primary, secondary=None):
        servers = [primary.ip]
        if secondary:
            servers.append(secondary.ip)
        return self._configure_param(_api,
            category="mgmt",
            param="environment.dns_servers",
            value=",".join(servers),
            subject="DNS server" + ("s" if len(servers) > 1 else ""),
            value_name=", ".join(servers),
            )

    @command(
        description="Set system name",
        types=(Name,),
        )
    @roles.INFINIDAT.ADMIN
    def set_name(self, _api, name):
        current_name = _api.get("system/name")
        if current_name == name:
            return 'System name is already "{}"'.format(name)
        _api.put("system/name", name)
        return 'System renamed from "{}" to "{}"'.format(current_name, name)

    @command(
        description="Enables / disables the redirection from HTTP (80)to HTTPS (443)",
        types=(YesNo,),
        )
    @roles.INFINIDAT.ADMIN.TECHNICIAN
    def set_ssl_redirection(self, _api, redirect):
        system = _api.get(self._BASE_URI)
        if system.mgmt["network.force_ssl_all"] == redirect :
            return "HTTPS redirection is already {}".format("enabled" if redirect else "disabled")
        return self._configure_param(_api,
            category="mgmt",
            param="network.force_ssl_all",
            value=redirect,
            subject="HTTPS redirection",
            value_name="{}".format("enabled" if redirect else "disabled"),
        )

    @command(
        description="Set system MOTD",
        types=(Str,),
        )
    @roles.INFINIDAT.ADMIN.TECHNICIAN
    def set_motd(self, _api, motd):
        system = _api.get(self._BASE_URI)
        if system.mgmt["client.message_of_the_day"] == motd :
            return "System motd is alreasy set to:\n{}".format(motd)
        return self._configure_param(_api,
            category="mgmt",
            param="client.message_of_the_day",
            value=motd,
            subject="System motd set to:",
            value_name="{}".format(motd),
        )


@commandgroup(parent=SystemConfigCommands, prefix="time")
class SystemTimeCommands(ConfigCommandsBase):

    @command(
        description="Set NTP servers",
        types=(Hostname, Hostname),
        )
    @roles.INFINIDAT.ADMIN
    def set_ntp(self, _api, primary, secondary=None):
        servers = [primary]
        if secondary:
            servers.append(secondary)
        return self._configure_param(_api,
            category="mgmt",
            param="environment.ntp_servers",
            value=",".join(servers),
            subject="NTP server" + ("s" if len(servers) > 1 else ""),
            value_name=", ".join(servers),
            )
