from functools import partial
from munch import Munch

from infi.shell import commandgroup, command
from infi.shell.types import Str, YesNo, Int
from infi.shell.exceptions import BadArguments, BadValue
from infi.shell.utils import strlist, NOTHING

from .filesystem import FilesystemCommands, FilesystemCommandsMixin
from .types import (
    Size,
    Export,
    ExportClient,
    ExportAccessLevel,
    Filesystem,
    )
from .options import QueryOptions
from . import (
    EntityCommands,
    ENTITY,
    roles,
    render_collection,
    LazyEntityMapping,
    will_render_any_field,
    )

class FilesystemExportCommandsBase(EntityCommands):

    def _get_lazy_fs_mapping(self, api, ids=None, preloaded=None):
        filter = None if ids is None else api.create_filter().member_of("id", *ids)
        return LazyEntityMapping(Filesystem, api, fields=("id", "name", "type"), filter=filter, preloaded_entities=preloaded)

@commandgroup(parent=FilesystemCommands, prefix="export")
class FilesystemExportCommands(FilesystemCommandsMixin, FilesystemExportCommandsBase):
    ENTITY_TYPE = Export
    QUERY_FIELDS = strlist("""
        export_path
        inner_path
        filesystem
        filesystem_type
        enabled
        permissions_count
        """)
    DETAILED_QUERY_FIELDS = QUERY_FIELDS + strlist("""
        anonymous_gid
        anonymous_uid
        make_all_users_anonymous
        max_write
        max_read
        pref_write
        pref_read
        pref_readdir
        privileged_port
        32bit_file_id
        """)
    QUERY_SUBSTITUTES = dict(
        inner_path="internal_path",
        filesystem_type="type",
        permissions_count="permissions",
        anonymous_gid="anongid",
        anonymous_uid="anonuid",
        make_all_users_anonymous="squash_all_users",
        max_write="max_wsize",
        max_read="max_rsize",
        pref_write="pref_wsize",
        pref_read="pref_rsize",
        pref_readdir="pref_readdir_size",
        privileged_port="privileged_ports_only",
        )
    SORTABLE_FIELDS = DETAILED_QUERY_FIELDS - strlist("""
        filesystem
        filesystem_type
        """)

    @command(
        description="List existing filesystem exports",
        types=(Export.List, Str.List, Filesystem.List),
        substitute=dict(
            export="export_path",
            fs="name",
            ),
        options=QueryOptions(fields=DETAILED_QUERY_FIELDS, sortable=SORTABLE_FIELDS, substitute=QUERY_SUBSTITUTES, detailed=True),
        )
    @render_collection(
        entity=Export,
        fields=QUERY_FIELDS,
        substitute=QUERY_SUBSTITUTES,
        detailed_individually=True,
        detailed_fields=DETAILED_QUERY_FIELDS,
        )
    def query(self, _api, export=None, internal_path=None, fs=None):
        filesystem_by_id = self._get_lazy_fs_mapping(_api, preloaded=fs) if fs else None
        return self._query(_api, export,
            inner_path=internal_path,
            filesystem_id=fs,
            _hook=partial(self._query_hook,
                api=_api,
                filesystem_by_id=filesystem_by_id,
                with_fs_info=will_render_any_field("filesystem", "filesystem_type"),
                ),
            )

    def _query_hook(self, exports, api=None, filesystem_by_id=None, with_fs_info=False):
        if with_fs_info:
            if not filesystem_by_id and api:
                filesystem_by_id = self._get_lazy_fs_mapping(api, ids=(e.filesystem_id for e in exports))
            if filesystem_by_id:
                for e in exports:
                    fs = filesystem_by_id[e.filesystem_id]
                    e.update(
                        filesystem=fs.name,
                        filesystem_type=self._get_fs_type(fs),
                        permissions_count=len(e.permissions),
                        )
        return exports

    def _generate_export_name(self, export_path):
        return export_path.strip("/").replace("/", "_")

    @command(
        description="Create a new filesystem export",
        types=(Filesystem, Str, Str, Export,
               YesNo, Int, Int,
               Size, Size,
               Size, Size, Size,
               YesNo),
        substitute=dict(
            permissions_export="copy_permissions_from",
            ),
        )
    @roles.INFINIDAT.ADMIN.POOLADMIN
    def create(self, _api, fs, export_path, internal_path=None, permissions_export=None,
                           squash_all_users=None, anonymous_gid=None, anonymous_uid=None,
                           maximum_write_size=None, maximum_read_size=None,
                           preferred_write_size=None, preferred_read_size=None, preferred_readdir_size=None,
                           privileged_port=None):
        props = self._create_export_props(**locals())
        props.update(
            filesystem_id=fs.id,
            export_path=export_path,
            inner_path=internal_path,
            )
        return self._create(_api, **props)

    def _create_export_props(_self, permissions=None, permissions_export=None,
                                    squash_all_users=None, anonymous_gid=None, anonymous_uid=None,
                                    maximum_write_size=None, maximum_read_size=None,
                                    preferred_write_size=None, preferred_read_size=None, preferred_readdir_size=None,
                                    privileged_port=None,
                                    **__):
        if permissions and permissions_export:
            raise BadArguments("Please specify either permissions or copy_permissions_from")
        props = Munch(
            make_all_users_anonymous=squash_all_users,
            anonymous_gid=anonymous_gid,
            anonymous_uid=anonymous_uid,
            max_write=maximum_write_size,
            max_read=maximum_read_size,
            pref_write=preferred_write_size,
            pref_read=preferred_read_size,
            pref_readdir=preferred_readdir_size,
            privileged_port=privileged_port,
            )
        if permissions:
            _self._apply_permissions_on_export_props(props, permissions)
        elif permissions_export:
            props.permissions = permissions_export.permissions
        return props

    def _apply_permissions_on_export_props(self, props, permissions):
        perms = []
        for acl in permissions:
            perms += (Munch(access=acl,client=cli) for cli in permissions[acl].split(',') if cli != "")
        props['permissions'] = list(perms)

    @command(
        description="Modify existing filesystem export",
        types=(Export, Export,
               YesNo, Int, Int,
               Size, Size,
               Size, Size, Size,
               YesNo),
        substitute=dict(
            export="export_path",
            permissions_export="copy_permissions_from",
            ),
        )
    @roles.INFINIDAT.ADMIN.POOLADMIN
    def modify(self, _api, export, permissions_export=None,
                           squash_all_users=None, anonymous_gid=None, anonymous_uid=None,
                           maximum_write_size=None, maximum_read_size=None,
                           preferred_write_size=None, preferred_read_size=None, preferred_readdir_size=None,
                           privileged_port=None):
        props = self._create_export_props(**locals())
        if not any(value is not None for value in props.itervalues()):
            raise BadArguments("At least one more argument must be specified")
        self._update(_api, export, **props)
        return self._result_message(export, "modified")

    @command(
        description="Enable filesystem export",
        types=(Export,),
        substitute=dict(
            export="export_path",
            ),
        )
    @roles.INFINIDAT.ADMIN.POOLADMIN
    def enable(self, _api, export):
        if export.enabled:
            return self._result_message(export, "is already enabled")
        self._update(_api, export, enabled=True)
        return self._result_message(export, "enabled")

    @command(
        description="Disable filesystem export",
        types=(Export,),
        substitute=dict(
            export="export_path",
            ),
        )
    @roles.INFINIDAT.ADMIN.POOLADMIN
    def disable(self, _api, export):
        if not export.enabled:
            return self._result_message(export, "is already disabled")
        self._update(_api, export, enabled=False)
        return self._result_message(export, "disabled")

    @command(
        description="Delete filesystem export",
        types=(Export.List,),
        substitute=dict(
            export="export_path",
            ),
        )
    @roles.INFINIDAT.ADMIN.POOLADMIN
    def delete(self, _api, export):
        return self._delete(_api, export)


@commandgroup(parent=FilesystemExportCommands, prefix="permission")
class FilesystemExportPermissionCommands(FilesystemCommandsMixin, FilesystemExportCommandsBase):
    ENTITY_TYPE = Export
    QUERY_FIELDS = strlist("""
        export_path
        client
        access
        no_root_squash
        """)
    DETAILED_QUERY_FIELDS = strlist("""
        export_path
        filesystem
        filesystem_type
        client
        access
        no_root_squash
        """)
    QUERY_SUBSTITUTES = dict(
        access="access_type",
        filesystem_type="type",
        )
    SORTABLE_FIELDS = strlist("export_path")

    @command(
        description="List permissions for existing filesystem exports",
        types=(Export.List, Filesystem.List, ExportClient.List, ExportAccessLevel.List, YesNo),
        substitute=dict(
            exports="export_path",
            fs="name",
            ),
        options=QueryOptions(fields=DETAILED_QUERY_FIELDS, sortable=SORTABLE_FIELDS, substitute=QUERY_SUBSTITUTES),
        )
    @render_collection(
        entity=Export,
        fields=QUERY_FIELDS,
        substitute=QUERY_SUBSTITUTES,
        detailed_individually=True,
        detailed_fields=DETAILED_QUERY_FIELDS,
        )
    def query(self, _api, exports=None, fs=None, client=None, access=None, no_root_squash=None):
        filesystem_by_id = self._get_lazy_fs_mapping(_api, preloaded=fs) if fs else None
        return self._query(_api, exports,
            _hook=partial(self._query_hook,
                api=_api,
                client=client,
                access=access,
                fs=fs,
                no_root_squash=no_root_squash,
                filesystem_by_id=filesystem_by_id,
                with_fs_info=will_render_any_field("filesystem", "filesystem_type"),
                ),
            )

    def _query_hook(self, exports, api=None, client=None, access=None, fs=None, no_root_squash=None,
                    filesystem_by_id=None, with_fs_info=False):
        if with_fs_info and not filesystem_by_id:
            assert api
            filesystem_by_id = self._get_lazy_fs_mapping(api, ids=(e.filesystem_id for e in exports))
        permissions = self._query_iter_permissions(exports, client, access, fs, filesystem_by_id, with_fs_info)
        return self._filter_entities(permissions, client=client, access=access, no_root_squash=no_root_squash)

    def _query_iter_permissions(self, exports, client, access, fs, filesystem_by_id, with_fs_info):
        if fs:
            exports = [export for export in exports for f in fs if export.filesystem_id == f.id]
        for export in exports:
            for permission in sorted(export.permissions, key=lambda _p: _p.client):
                permission.update(
                    export_path=export.export_path,
                    access=ExportAccessLevel.MGMT_LEVELS[permission.access],
                    )
                if with_fs_info:
                    fs = filesystem_by_id[export.filesystem_id]
                    permission.update(
                        filesystem=fs.name,
                        filesystem_type=self._get_fs_type(fs),
                        )
                yield permission

    #     permissions = (
    #         self._create_permissions_object(export, permission, filesystem_by_id)
    #         for export in exports
    #         for permission in sorted(export.permissions, key=lambda _p: _p.client)
    #         if permission.client
    #         )
    #     return self._filter_entities(permissions, client=client, access=access)

    # def _create_permissions_object(self, export, permission, filesystem_by_id):
    #     obj = Munch(
    #         export_path=export.export_path,
    #         client=permission.client,
    #         access=ExportAccessLevel.MGMT_LEVELS[permission.access],
    #         no_root_squash=permission.no_root_squash,
    #         )
    #     if filesystem_by_id:
    #         fs = filesystem_by_id[export.filesystem_id]
    #         obj.update(
    #             filesystem=fs.name,
    #             filesystem_type=self._get_fs_type(fs),
    #             )
    #     return obj

    def _find_permission(self, permissions, client):
        for p in permissions:
            if p.client == client:
                return p
        return None

    @command(
        description="Add a new permission to an existing filesystem export",
        types=(Export, ExportClient, ExportAccessLevel, YesNo),
        substitute=dict(
            export="export_path",
            ),
        )
    @roles.INFINIDAT.ADMIN.POOLADMIN
    def add(self, _api, export, client, access, no_root_squash=None):
        permissions = export.permissions
        if self._find_permission(permissions, client):
            self._raise_invalid_client(export, client, True)
        permission = self._filter_properties(dict(
            client=client,
            access=access,
            no_root_squash=no_root_squash,
            ))
        permissions.append(permission)
        result = self._update(_api, export, permissions=permissions, returns=ENTITY)
        permission = self._find_permission(result.permissions, client) or permission
        return self._permission_result_message("added to", export, permission)

    @command(
        description="Remove a permission from an existing filesystem export",
        types=(Export, ExportClient),
        substitute=dict(
            export="export_path",
            ),
        )
    @roles.INFINIDAT.ADMIN.POOLADMIN
    def remove(self, _api, export, client):
        permissions = export.permissions
        permission = None
        for i, p in enumerate(permissions):
            if p.client == client:
                permission = p
                permissions.pop(i)
                break
        if permission is None:
            self._raise_invalid_client(export, client, False)
        self._update(_api, export, permissions=permissions)
        return self._permission_result_message("removed from", export, permission)

    @command(
        description="Add a new permission to an existing filesystem export",
        types=(Export, ExportClient, ExportAccessLevel, YesNo),
        substitute=dict(
            export="export_path",
            ),
        )
    @roles.INFINIDAT.ADMIN.POOLADMIN
    def modify(self, _api, export, client, access=None, no_root_squash=None):
        if all(attr is None for attr in (access, no_root_squash)):
            raise BadArguments("At least one more argument must be specified")
        permissions = export.permissions
        permission = self._find_permission(permissions, client)
        if permission is None:
            self._raise_invalid_client(export, client, False)
        updated = False
        for attr, value in (
            ("access", access),
            ("no_root_squash", no_root_squash),
            ):
            if value is not None:
                permission[attr] = value
                updated = True
        if updated:
            self._update(_api, export, permissions=permissions)
        return self._permission_result_message("modified on", export, permission)

    def _permission_result_message(self, action, export, permission):
        result = 'Client "{}" {} {} ({})'.format(
            permission.client,
            action,
            self._result_subject_and_entity(export),
            self._humanize_permission_properties(permission),
            )
        return result

    def _humanize_permission_properties(self, p):
        parts = []
        for display_name, attr, transformer in (
            ("access",          "access",           lambda value: ExportAccessLevel.MGMT_LEVELS[value]),
            ("no root squash",  "no_root_squash",   lambda value: "yes" if value else "no")
            ):
            value = p.get(attr, NOTHING)
            if value is NOTHING:
                continue
            if transformer:
                value = transformer(value)
            parts.append("{}: {}".format(display_name, value))
        return ", ".join(parts)

    def _raise_invalid_client(self, export, client, is_defined):
        raise BadValue('Client "{}" is {} defined for {}'
                       .format(client, "already" if is_defined else "not", self._result_subject_and_entity(export)))
