from functools import partial
from munch import Munch
from itertools import chain

from infi.shell import commandgroup, command, render, config
from infi.shell.utils import strlist

from .types import Volume, Host, Cluster
from .options import QueryOptions
from . import set_renderer_query_fields

from ..renderer import render_capacity

class MappingCommandsBase(object):
    """
    Base class for command groups which require volume mapping capabilities.
    """

    def _map(self, _api, mappable_type, mappable, volumes, lun):
        aggregator = _api.create_request_aggregator()
        with aggregator:
            for volume in volumes:
                aggregator.next_task(volume.name, partial(self._get_mapping_success_message,
                    is_mapping=True,
                    volume=volume,
                    mappable_type=mappable_type,
                    mappable=mappable,
                    ))
                props = {"volume_id" : volume.id}
                if lun is not None:
                    props["lun"] = lun
                _api.post(mappable_type.get_uri(mappable.id, "luns"), props)
        return aggregator.run()

    def _unmap(self, _api, mappable_type, mappable, volumes):
        aggregator = _api.create_request_aggregator()
        with aggregator:
            for volume in volumes:
                aggregator.next_task(volume.name, partial(self._get_mapping_success_message,
                    is_mapping=False,
                    volume=volume,
                    mappable_type=mappable_type,
                    mappable=mappable,
                    ))
                _api.delete(mappable_type.get_uri(mappable.id, "luns", "volume_id", volume.id))
        return aggregator.run()

    def _unmap_lun(self, _api, mappable, lun):
        result = _api.delete(self.ENTITY_TYPE.get_uri(mappable.id, "luns", "lun", lun))
        volume = Volume.get_by_id(_api, result.volume_id)
        return self._get_mapping_success_message(
            result=result,
            is_mapping=False,
            volume=volume,
            mappable_type=self.ENTITY_TYPE,
            mappable=mappable,
            )

    def _get_mapping_success_message(self, result, is_mapping, volume, mappable_type, mappable):
        return '{volume_type} "{volume_name}" {action} LUN {lun} in {mappable_type} "{mappable.name}"'.format(
            volume_type=Volume.mgmt_subtype_to_class(volume.type).get_entity_name().capitalize(),
            volume_name=volume.name,
            action="mapped to" if is_mapping else "unmapped from",
            lun=result.lun,
            mappable_type=mappable_type.get_entity_name(),
            mappable=mappable,
            )

    def _general_map_query(self, _api, lun_containers):
        return [Munch({self.ENTITY_TYPE.get_entity_name() : lc.name,
                       "lun" : mapping.lun,
                       "volume" : Volume.get_by_id(_api, mapping.volume_id).name})
                for lc in lun_containers
                for mapping in lc.luns]

QUERY_FIELDS = strlist("name type lun volume size")

QUERY_SUBSTITUTES = dict(
    name="target_name",
    type="target_type",
    )

@commandgroup(prefix="map")
class MappingCommands(MappingCommandsBase):
    @command(
        description="List existing host and cluster mappings",
        options=QueryOptions(fields=QUERY_FIELDS, substitute=QUERY_SUBSTITUTES),
        )
    @render(
        fields=QUERY_FIELDS,
        substitute=QUERY_SUBSTITUTES,
        )
    def query(self, _api):
        set_renderer_query_fields()
        # The actual query is separated to a function of its own, since that
        # function is a generator, and the code above needs to be executed
        # during the command's run time, and not during the rendering. All the
        # rest can be executed during rendering.
        return self._query_generator(_api)

    def _query_generator(self, _api):
        volume_ids = set()
        def host_query_hook(_hosts):
            for _host in _hosts:
                if _host.host_cluster_id:
                    _host.luns[:] = [lun for lun in _host.luns if not lun.clustered]
                if _host.luns:
                    for info in query_hook((_host,), Host):
                        yield info
        def query_hook(_mappables, entity_type):
            for m in _mappables:
                for lun in m.luns:
                    volume_ids.add(lun.volume_id)
                    yield Munch(
                        name=m.name,
                        type=entity_type.get_entity_name().upper(),
                        lun=lun.lun,
                        volume=lun.volume_id,  # Store volume ID, which will be replaced by name later
                        )
        # Get hosts and clusters
        pluck = _api.create_pluck("name", "luns")
        hosts = Host.iter_all(_api, pluck=pluck.pluck("host_cluster_id"), hook=host_query_hook, page_size=config.api_max_page_size)  # Assume only 1000 hosts
        clusters = Cluster.iter_all(_api, pluck=pluck, hook=partial(query_hook, entity_type=Cluster), page_size=config.api_max_page_size)  # Assume only 1000 clusters
        # Arrange everything and sort
        all_mappables = sorted(chain(hosts, clusters), key=lambda m: m.name)
        # If no volumes are mapped, there's no point in proceeding
        if not volume_ids:
            return
        # Get all relevant volumes in one query
        volume_ids = list(volume_ids)
        vol_page_size = 100  # Restrict volumes page size to avoid excessively long URLs in this particular query
        volume_names_by_id = {}
        volume_sizes_by_id = {}
        for index in xrange(0, len(volume_ids), vol_page_size):
            for v in Volume.iter_all(_api, filter=_api.create_filter().member_of("id", *volume_ids[index:index+100]), pluck=_api.create_pluck("id", "name", "size"), page_size=vol_page_size):
                volume_names_by_id[v.id] = v.name
                volume_sizes_by_id[v.id] = v.size
        # Replace volume IDs with volume names, add size, and return results
        for m in all_mappables:
            m.size = volume_sizes_by_id[m.volume]
            m.volume = render_capacity(volume_names_by_id[m.volume])
            yield m
