import capacity
import re
import os
import itertools
import string
from munch import Munch
from numbers import Number
from collections import OrderedDict

from infi.shell import config
from infi.shell.types import (
    ArgType,
    ListCreator,
    Int,
    Str,
    Range as _Range,
    Positive,
    Measure,
    RegexType,
    ChoiceType,
    MappedChoiceType,
    Dynamic,
    TypeChoicesHelpMixin,
    IPAddress,
    TimeDelta,
    )
from infi.shell.utils import decamelize, SpellingCorrector
from infi.shell.exceptions import BadValue

from ...utils import capitalize as _capitalize, NamedOrderedDict
from ..exceptions import NotFound, ManagementException, NoSuchEntity, InvalidConstant
from ..api import ParamFactory, QueryFilter, QuerySort
from ..system import system
from ..system.user import Roles


NONE_CONSTANT = "__none__"



################################################################################
#  PATCHES AND SETTINGS                                                        #
################################################################################

ChoiceType.AUTO_COMPLETION_CASE_INSENSITIVE = True
MappedChoiceType.AUTO_COMPLETION_CASE_INSENSITIVE = True



################################################################################
#  UTILITIES                                                                   #
################################################################################

def combine_type_names(*types):
    return combine_names(*(t.get_name() for t in types))

def combine_names(*names):
    return "{{{}}}".format("|".join(names))

class ConstantsHelpMixin(TypeChoicesHelpMixin):
    """ Mixin for automatically generating name and description for subtypes
    which override ``CONSTANTS``. """
    def get_description(self):
        desc = super(ConstantsHelpMixin, self).get_description()
        return "{}, or {}{}".format(", ".join(self.CONSTANTS), desc[0].lower(), desc[1:])
    def get_type_name_parts(self):
        return list(self._constants) + [self.get_name()]

class SubtypeQueryFilter(object):
    """ Generic ``QueryFilter`` substitute for entity hierarchies. An entity
    with subtypes may declare an instance of this class in its ``QUERY_FILTER``
    class attribute, specifying which entity field should be filtered, and in
    which class attribute the value is held. """
    def __init__(self, filtered_field, subtype_attr):
        self.filtered_field = filtered_field
        self.subtype_attr = subtype_attr
    def __get__(self, owner, type):
        assert hasattr(type, self.subtype_attr), \
            '{} subtype query filter expects class attribute "{}", but no such attribute is defined' \
            .format(type.__name__, self.subtype_attr)
        value = getattr(type, self.subtype_attr)
        if value:
            return QueryFilter().eq(self.filtered_field, value)
        return None



################################################################################
#  SIMPLE TYPES                                                                #
################################################################################


class HostMappingType(MappedChoiceType):
    MAPPING = dict(
        HOST=False,
        CLUSTER=True,
        )
    def _get_mapping(self):
        return self.MAPPING


class Range(_Range, TypeChoicesHelpMixin):
    TYPE_NAME_SEPARATOR = ".."
    DEFAULT_VALUE = None  # Note that the default value has to be used manually in the command
    def get_type_name_parts(self):
        return (str(self.MIN), str(self.MAX))
    def get_description(self):
        description = super(Range, self).get_description()
        return description if self.DEFAULT_VALUE is None \
               else "{} (default: {})".format(description, self.DEFAULT_VALUE)

class Regex(ArgType):
    def _get_actual(self, pattern):
        try:
            return True, re.compile(pattern, re.I)
        except Exception as e:
            return False, "contains an invalid regular expression: {}".format(e)
    @classmethod
    def get_name(cls):
        return "PATTERN"
    def get_description(self):
        return "A regular expression pattern (case insensitive)"

class Name(RegexType):
    MAX_LENGTH = 65
    _ALLOWED_SYMBOLS = "^&'@()[]$=!-#{}%.+~_ "
    ALLOWED_CHARACTERS = set(string.digits + string.letters + _ALLOWED_SYMBOLS)
    @classmethod
    def _filter_func(cls, c):
        return c in cls.ALLOWED_CHARACTERS
    @classmethod
    def filter(cls, s):
        return filter(cls._filter_func, s)
    def _get_pattern(self):
        return r"^[0-9a-zA-Z{}]*$".format(re.escape(self._ALLOWED_SYMBOLS))
    def _verify_raw(self, value):
        if not (value and value.strip()):
            return "is empty"
        return super(Name, self)._verify_raw(value)
    def _get_error(self):
        return "contains illegal characters"
    def get_description(self):
        return "A maximum of {} Latin characters, numbers, spaces, and the following symbols: " \
                "\"{}\" (excluding quotation marks). Leading and trailing whitespace " \
                "characters are stripped.".format(self.MAX_LENGTH, self._ALLOWED_SYMBOLS.strip())

class NetworkInterfaceName(Name):
    MAX_LENGTH = 9

class Series(RegexType):
    """ A series of numbers, with optional padding. The actual value is a
    ``SeriesDescriptor``. """
    MAX_VALUE = 10000000000
    MAX_AMOUNT = MAX_VALUE
    class SeriesDescriptor(object):
        def __init__(self, start, end, padding):
            self.start, self.end = start, end
            self.padding = padding
            self.total = end + 1 - start
        def generator(self, name, padding=None):
            formatter = name + "{{:0{}}}".format(self.padding if padding is None else padding)
            iterator = xrange(self.start, self.end+1) if self.start <= self.end else xrange(self.end, self.start-1, -1)
            return (formatter.format(n) for n in iterator)
        def __repr__(self):
            return "<{t} start={s.start} end={s.end} total={s.total} padding={s.padding}>" \
                   .format(t=type(self).__name__, s=self)
    def _get_pattern(self):
        return r"((?P<start>\d+)-)?(?P<end>\d+)"
    def _get_actual(self, value):
        match = self._get_match(value)
        start_str, end_str = match.group("start"), match.group("end")
        no_start = not start_str
        start_str = start_str or "1"
        start, end = int(start_str), int(end_str)
        padding = max((len(start_str), len(end_str)))
        if no_start and end == 0:
            return False, "is an empty series"
        if end < start:
            start, end = end, start
        descriptor = self.SeriesDescriptor(start, end, padding)
        if descriptor.end > self.MAX_VALUE:
            return False, "last item exceeds the allowed maximum ({})".format(self.MAX_VALUE)
        if descriptor.total > self.MAX_AMOUNT:
            return False, "has {} items, but the allowed maximum is {}".format(descriptor.total, self.MAX_AMOUNT)
        return True, descriptor
    def _get_error(self):
        return "is not a valid sequence"
    def get_description(self):
        return 'Either a range of numbers, e.g "3-7", or a single number, e.g "5", which ' \
               'is the same as "1-5". Zero-padding the numbers yields a zero-padded sequence. ' \
               'The series may contain at most {} items, and the index of the last item may ' \
               'not exceed {}'.format(self.MAX_AMOUNT, self.MAX_VALUE)

class VolumeSeries(Series):
    MAX_AMOUNT = 100
    @classmethod
    def get_name(cls):
        return 'SERIES'

class FSSeries(Series):
    MAX_AMOUNT = 100
    @classmethod
    def get_name(cls):
        return 'SERIES'

class SCSIPortType(ChoiceType):
    TYPES = Munch(
        FC="fc",
        ISCSI="iscsi",
        )
    def _get_values(self):
        return self.TYPES.values()
    @classmethod
    def get_name(cls):
        return "TYPE"

class WWN(RegexType):
    PATTERN = r"([0-9a-fA-F]{2}:?){7}[0-9a-fA-F]{2}"
    def _get_pattern(self):
        return self.PATTERN
    def _get_error(self):
        return "is not a valid WWN"
    def get_description(self):
        return "A 16-digit hexadecimal number, optionally divided by colons to 8 groups of two digits each"

CAPACITY_UNITS = OrderedDict((unit, cap)
              for cap, units in (
                  (capacity.byte, ("B")),
                  (capacity.KB, ("K", "KB")),
                  (capacity.KiB, ("KiB",)),
                  (capacity.MB, ("M", "MB")),
                  (capacity.MiB, ("MiB",)),
                  (capacity.GB, ("G", "GB")),
                  (capacity.GiB, ("GiB",)),
                  (capacity.TB, ("T", "TB")),
                  (capacity.TiB, ("TiB",)),
                  (capacity.PB, ("P", "PB")),
                  (capacity.PiB, ("PiB",)),
                  )
              for unit in units)

BLOCK_CAPACITY = capacity.byte * 512  # Infinibox block size

class Size(Measure):
    ROUNDUP_SIZE = BLOCK_CAPACITY
    def _is_float(self, value):
        return isinstance(value, Number) and long(value) != value
    def _verify(self, value):
        error = super(Size, self)._verify(value)
        if error:
            return error
        if self._is_float(value):
            return self._get_error()
    def _process_actual(self, size, unit):
        if not unit:
            return size if not self._is_float(size) else None
        unit = unit.lower()
        if unit == "b" and self._is_float(size):
            return None
        for u, cap in CAPACITY_UNITS.iteritems():
            if u.lower() == unit:
                return long((size * cap).roundup(self.ROUNDUP_SIZE) / capacity.byte)
        return None
    def _get_error(self):
        return "is not a valid size"
    def _is_negative_enabled(self):
        return False
    def _get_fraction_digits(self):
        return 2
    def _get_units(self):
        return CAPACITY_UNITS.iterkeys()
    def get_description(self):
        return "A positive number (up to 2 digits beyond the decimal point),"\
               " optionally followed by a unit; e.g. 1TB, 1000000000000"

class SizeDelta(Size):
    class SizeContainer(object):
        def __init__(self, size, sign):
            self._size, self._sign = size, sign
        def is_delta(self):
            return bool(self._sign)
        def get_size(self):
            assert not self.is_delta()
            return self._size
        def get_delta(self):
            assert self.is_delta()
            return self._size * (abs(self._sign) / self._sign)
        def compute_size(self, base_size):
            if self.is_delta():
                return base_size + self.get_delta()
            return self.get_size()
    def _get_actual(self, value):
        sign = None
        if value and value[:1] in ("+-"):
            sign, value = value[0], value[1:]
        is_valid, size = super(SizeDelta, self)._get_actual(value)
        if not is_valid:
            # "size" is the error message in this case
            return False, size
        return True, self.SizeContainer(size, 0 if sign is None else 1 if sign == "+" else -1)
    def _verify(self, value):
        return super(SizeDelta, self)._verify(value._size)
    def get_description(self):
        return "A number (up to 2 digits beyond the decimal point), optionally"\
               " followed by a unit; e.g. 1TB, 1000000000000. When preceded by"\
               " a plus or minus sign, represents a delta; e.g. 3000000m"\
               " (absolute size), +2000GB (positive delta), -1.5T (negative delta)"

class CapacityUnit(MappedChoiceType):
    def _get_mapping(self):
        mapping = OrderedDict((name, (capacity, name + ("" if name[-1] == "B" else "B")))
                              for name, capacity in CAPACITY_UNITS.iteritems()
                              if name[0] in "BGT")
        mapping['block'] = (BLOCK_CAPACITY, 'blocks')
        return mapping

    def get_description(self):
        return 'B, G (or GB), GiB, T (or TB), TiB or block'

class Lun(Int):
    def _verify(self, value):
        if value < 0:
            return "must not be negative"
    def get_description(self):
        return "SCSI logical unit number (LUN)"

class AutoExtend(ConstantsHelpMixin, Size):
    CONSTANTS = Munch(
        UNLIMITED=-1,
        DISABLED=0,
        )
    @classmethod
    def get_name(cls):
        return "SIZE"

class LdapDN(Str):
    def get_description(self):
        return "LDAP distinguished name"

class IPAddressOrNone(IPAddress):
    CONSTANTS = Munch(NONE=NONE_CONSTANT)

class TimeDeltaOrNone(TimeDelta):
    CONSTANTS = Munch(NONE=NONE_CONSTANT)
    @classmethod
    def get_name(cls):
        return "TIMEDELTA"

class MTU(Range):
    MIN = 600
    MAX = 9000
    def get_description(self):
        return "Maximum transmission unit. {}".format(super(MTU, self).get_description())


################################################################################
#  ENTITY BASE CLASSES                                                         #
################################################################################

class Entity(Dynamic):
    URI = None  # Override or implement get_uri() to prevent deducing URI from class name
    SEARCHABLE = True  # Whether or not type is supported by search API
    PAGED = True  # Whether or not pagination applies to this type
    PLUCKABLE = True
    FILTERABLE = True
    CASE_INSENSITIVE_NAME = False
    SORTABLE = True
    UNIQUE_NAME = True  # when False, argument value is a list
    MULTI_FIELD_PUT = False
    MULTI_FIELD_PUT_ALWAYS_ALL_FIELDS = False
    ID_FIELD = "id"
    NAME_FIELD = "name"
    TYPE_NAME = None  # Type display name (default: class name in upper case)
    ENTITY_NAME = None  # Entity display name (default: class name in lower case)
    ENTITY_SHORT_NAME = None  # Entity short name (not a display name) (default: entity name)
    ENTITY_MGMT_NAME = None  # Entity API name, for special queries (doesn't apply
                             # to all entity types; default: entity name)
    ENTITY_PLURAL_NAME = None  # Entity plural display name (default: semi-smart guess)
    DESCRIPTION = None  # Override to set entire description
    HOOK_DEPENDENCIES = None  # when default query hook is defined, this attribute
                              # must contain names of fields on which the hook depends
    QUERY_PARAMS = None  # Default query params (dict or any ParamFactory)
    QUERY_FILTER = None  # Default query filter (dict or QueryFilter)
    QUERY_SORT = None  # Default query sort (dict or QuerySort)
    List = ListCreator(allow_duplicates=False)
    def _verify_raw(self, value):
        if not (value and value.strip()):
            return "is empty"
        return super(Entity, self)._verify_raw(value)
    @classmethod
    def get_name(cls):
        if not hasattr(cls, "__type_name"):
            cls.__type_name = (getattr(cls, "TYPE_NAME", None) or super(Entity, cls).get_name()).upper()
        return cls.__type_name
    @classmethod
    def get_entity_name(cls, capitalize=False):
        if not hasattr(cls, "__entity_name"):
            cls.__entity_name = (getattr(cls, "ENTITY_NAME", None) or decamelize(cls.__name__)).lower()
        value = cls.__entity_name
        if capitalize:
            value = _capitalize(value)
        return value
    @classmethod
    def get_entity_short_name(cls, capitalize=False):
        if not hasattr(cls, "__entity_short_name"):
            cls.__entity_short_name = (getattr(cls, "ENTITY_SHORT_NAME", None) or cls.get_name().lower())
        value = cls.__entity_short_name
        if capitalize:
            value = _capitalize(value)
        return value
    @classmethod
    def get_entity_mgmt_name(cls):
        if not hasattr(cls, "__entity_mgmt_name"):
            cls.__entity_mgmt_name = (getattr(cls, "ENTITY_MGMT_NAME", None) or cls.__name__).lower()
        return cls.__entity_mgmt_name
    @classmethod
    def get_entity_plural_name(cls, capitalize=False):
        if not hasattr(cls, "__entity_plural_name"):
            cls.__entity_plural_name = getattr(cls, "ENTITY_PLURAL_NAME", None)
            if not cls.__entity_plural_name:
                name = cls.get_entity_name()
                last = name[-1]
                if last == 's':
                    name += 'es'
                elif last == 'y':
                    name = name[:-1] + 'ies'
                else:
                    name += 's'
                cls.__entity_plural_name = name
        value = cls.__entity_plural_name
        if capitalize:
            value = _capitalize(value)
        return value
    @classmethod
    def get_description(cls):
        if not hasattr(cls, "__description"):
            cls.__description = getattr(cls, "DESCRIPTION", None) or \
                                "Name of an existing {}".format(cls.get_entity_name())
        return cls.__description
    @classmethod
    def get_uri(cls, *subresource):
        if not hasattr(cls, "__uri"):
            cls.__uri = cls.URI or (cls.__name__.lower() + "s")
        if not subresource:
            return cls.__uri
        return "{}/{}".format(cls.__uri, "/".join(str(s) for s in subresource))
    def _get_dynamic(self, api, arg, value):
        return self._fetch(api, arg, value)
    @classmethod
    def _fetch(cls, api, arg, value):
        if arg.endswith("_id"):
            arg = arg[:-3]
            return arg, cls.get_by_id(api, value)
        return arg, cls.get_by_name(api, value)
    @classmethod
    def _get_query_params(cls):
        all_params = {}
        for params in (
            cls.QUERY_PARAMS,
            cls.QUERY_FILTER,
            cls.QUERY_SORT,
            ):
            if params:
                if isinstance(params, ParamFactory):
                    params = params.to_params()
                all_params.update(params)
        return all_params
    @classmethod
    def _compute_query_params(cls, *params):
        all_params = {}
        for p in itertools.chain((cls._get_query_params(),), params):
            if isinstance(p, ParamFactory):
                p = p.to_params()
            if p:
                all_params.update(p)
        return all_params
    @classmethod
    def _compute_query_hook(cls, hook, pluck):
        default_hook_compatible_with_pluck = cls._is_default_hook_compatible_with_pluck(pluck)
        if not hook:
            return cls._query_hook if default_hook_compatible_with_pluck else None
        elif not default_hook_compatible_with_pluck:
            return hook
        def wrapper_hook(results):
            results = cls._query_hook(results)
            results = hook(results)
            return results
        return wrapper_hook
    @classmethod
    def _is_default_hook_compatible_with_pluck(cls, pluck):
        if pluck and pluck._fields and cls.HOOK_DEPENDENCIES:
            fields = set(pluck._fields)
            for field in cls.HOOK_DEPENDENCIES:
                if field not in fields:
                    return False
        return True
    @classmethod
    def paged_query(cls, api, params=None, filter=None, sort=None, pluck=None, hook=None, **kwargs):
        params = cls._compute_query_params(params, filter, sort, pluck)
        hook = cls._compute_query_hook(hook, pluck)
        return api.paged_query(cls.get_uri(), params=params, hook=hook, **kwargs)
    @classmethod
    def query(cls, api, params=None, filter=None, sort=None, pluck=None, hook=None, **kwargs):
        params = cls._compute_query_params(params, filter, sort, pluck)
        hook = cls._compute_query_hook(hook, pluck)
        result = api.get(cls.get_uri(), params=params, **kwargs)
        if hook:
            result = hook(result)
        return result
    @classmethod
    def iter_all(cls, api, params=None, filter=None, sort=None, pluck=None, page_size=None, hook=None):
        if cls.PAGED:
            return cls.paged_query(api, params=params, filter=filter, sort=sort, pluck=pluck,
                                        page_size=page_size, hook=hook).iter_all_objects()
        else:
            return iter(cls.get_all(api, params=params, filter=filter, sort=sort, pluck=pluck, hook=hook))
    @classmethod
    def get_all(cls, api, params=None, filter=None, sort=None, pluck=None, hook=None):
        if cls.PAGED:
            return list(cls.iter_all(api, params=params, filter=filter, sort=sort, pluck=pluck,
                                          page_size=config.api_max_page_size, hook=hook))
        else:
            return cls.query(api, params=params, filter=filter, sort=sort, pluck=pluck, hook=hook)
    @classmethod
    def _query_hook(cls, results):
        return results
    @classmethod
    def search(cls, api, query="", page_size=None):
        assert cls.SEARCHABLE, cls.__name__ + " is not searchable"
        return api.paged_query(
            "search",
            page_size=page_size or config.api_default_search_page_size,
            params=dict(
                type=cls.get_entity_mgmt_name(),
                query=query,
                ),
            hook=cls._search_hook,
            )
    @classmethod
    def _search_hook(cls, results):
        return results
    @classmethod
    def get_by_id(cls, api, id):
        try:
            entity = api.get("{}/{}".format(cls.get_uri(), id))
            cls._validate_fetched_entity(entity)
            return entity
        except NotFound:
            raise NoSuchEntity(u'No such {}: ID {}'.format(cls.get_entity_name(), id))
    @classmethod
    def get_by_name(cls, api, name):
        if not (name and name.strip()):
            raise NoSuchEntity(u"No such {}: name is empty".format(cls.get_entity_name()))
        entity = cls._fetch_entity_by_name(api, name)
        if entity is not None:
            cls._validate_fetched_entity(entity)
            return entity
        raise NoSuchEntity(u'No such {}: "{}"'.format(cls.get_entity_name(), name))
    @classmethod
    def _fetch_entity_by_name(cls, api, name):
        filter = api.create_filter().eq(cls.NAME_FIELD, name) if cls.FILTERABLE else None
        entities = cls.iter_all(api, filter=filter)
        if entities is not None:
            if cls.CASE_INSENSITIVE_NAME:
                lower_name = name.lower()
                def predicate(entity):
                    entity_name = entity[cls.NAME_FIELD]
                    return entity_name and entity_name.lower() == lower_name
            else:
                predicate = lambda entity: str(entity[cls.NAME_FIELD]) == name
            if cls.UNIQUE_NAME:
                for entity in entities:
                    if predicate(entity):
                        return entity
            else:
                entities = [entity for entity in entities if predicate(entity)]
                if entities:
                    return entities
        return None
    @classmethod
    def _validate_fetched_entity(cls, entity):
        """ Subclasses may implement this method, which is invoked during
        ``get_by_name()``. The method should check whether the fetched entity is
        valid, and throw ``NoSuchEntity`` in case it isn't. """
        pass
    @classmethod
    def get_single(cls, api, **iter_all_kwargs):
        entities = cls.iter_all(api, **iter_all_kwargs)
        entity = None
        for e in entities:
            if entity is not None:
                raise NoSuchEntity("Expected a single {}, but multiple {} were found"
                                   .format(cls.get_entity_name(), cls.get_entity_plural_name()))
            entity = e
        if entity is None:
            raise NoSuchEntity("No such {}".format(cls.get_entity_name()))
        return e

    def get_possible_values(self, api, starts_with=None, type_cls=None):
        type_cls = type_cls or type(self)
        if type_cls.SEARCHABLE:
            return self._get_possible_values_by_search(api, starts_with, type_cls)
        return self._get_possible_values_by_query(api, starts_with, type_cls)

    def _get_possible_values_by_search(self, api, starts_with, type_cls):
        starts_with = starts_with or ""
        query = api.paged_query(
            "search",
            page_size=config.api_max_page_size,
            params=dict(
                type=type_cls.get_entity_mgmt_name(),
                query=starts_with,
                ),
            hook=self._search_hook,
            )
        return (entity_info.name for entity_info in query if entity_info and entity_info.name.startswith(starts_with))

    def _get_possible_values_by_query(self, api, starts_with, type_cls):
        type_cls = type_cls or type(self)
        params = {}
        # Create pluck for entity's name field, and any other fields possibly necessary for invoking the default query hook
        pluck = None if not type_cls.PLUCKABLE else api.create_pluck(type_cls.NAME_FIELD, *(type_cls.HOOK_DEPENDENCIES or ()))
        # Create generator - note that this has to be a generator because params may be updated below
        generator = (e for e in type_cls.iter_all(api, params=params, pluck=pluck, page_size=config.api_max_page_size))
        if starts_with:
            if type_cls.FILTERABLE:
                params.update(api.create_filter().like(type_cls.NAME_FIELD, starts_with + "%").to_params())
            else:
                generator = (e for e in generator if e[type_cls.NAME_FIELD].startswith(starts_with))
        # Return only names, not whole entities
        return (e[type_cls.NAME_FIELD] for e in generator)

class ConstantEntity(Entity):
    _subclasses = []
    class __metaclass__(getattr(Entity, '__metaclass__', type)):
        def __new__(mcls, name, bases, attrs):
            new_class = getattr(Entity, '__metaclass__', type).__new__(mcls, name, bases, attrs)
            if name != "ConstantEntity":
                ConstantEntity._subclasses.append(new_class)
            return new_class
    AUTO_COMPLETION_CASE_INSENSITIVE = True
    CACHEABLE = True
    FETCH_ONCE = True
    SEARCHABLE = False
    PAGED = False
    URI = None # This must be overridden
    ATTR = None # This may be overridden
    FIELD = None  # This may be overridden
    _values = None
    _spelling_corrector = None
    @classmethod
    def _request_values(cls, api):
        params = api.create_pluck(cls.ATTR).to_params() if cls.ATTR else None
        values = api.get(cls.URI, params=params)
        if cls.ATTR:
            for attr in cls.ATTR.split('.'):
                values = getattr(values, attr)
        if cls.FIELD:
            values = [getattr(v, cls.FIELD) for v in values]
        return list(cls._transform_values(values))
    @classmethod
    def _transform_values(cls, values):
        return values
    @classmethod
    def _ensure_values(cls, api):
        if cls._values is None:
            values = cls._request_values(api)
            values.sort()
            cls._values = values
            cls._spelling_corrector = SpellingCorrector(cls._values)
    @classmethod
    def invalidate_cache(cls):
        """ Invalidate cache of constant values, allowing them to be re-fetched
        from server. """
        cls._values = None
    @classmethod
    def invalidate_all_caches(cls):
        """ Invalidate cache of constant values for all constants, allowing them
        to be re-fetched from server. This method is present only in
        ConstantEntity, and absent from all its subclasses. """
        for subclass in cls._subclasses:
            subclass.invalidate_cache()
    @classmethod
    def _fetch(cls, api, arg, value):
        cls._ensure_values(api)
        for const_value in cls._values:
            if const_value.lower() == value.lower():
                return arg, const_value
        suggestions = cls._spelling_corrector.correct(value)
        msg = u'No such {}: "{}"'.format(cls.get_entity_name(), value)
        if suggestions:
            msg += u". Did you mean: {}?".format(u", ".join(suggestions))
        raise InvalidConstant(msg)
    @classmethod
    def iter_all(cls, api):
        cls._ensure_values(api)
        return iter(cls._values)
    @classmethod
    def get_all(cls, api):
        return list(cls.iter_all(api))
    def get_possible_values(self, api, starts_with=None, type_cls=None):
        it = self.iter_all(api)
        if starts_with:
            it = (value for value in it if value.lower().startswith(starts_with.lower()))
        return it


class ComponentNestedEntity(Entity):
    SEARCHABLE = False
    FILTERABLE = False
    PAGED = False
    CONTAINER_TYPE = None  # MANDATORY: Subclasses must override this (Entity)
    CONTAINER_SUBCOMPONENTS_FIELD = None  # MANDATORY: Subclasses must override this (str)
                                          #            The field in the container, where the subcomponents are nested
    SUBCOMPONENT_CONTAINER_ID_FIELD = None  # MANDATORY: Subclasses must override this (str)
                                            #            The field in the subcomponent, used to identify its container
    NAME_PATTERN = None  # MANDATORY: Subclasses must override this (compiled regex)
    ENTITY_NAME_FORMATTER = None  # MANDATORY: Subclasses must override this (str, suitable for str.format())
    ENTITY_NAME_PART_FIELDS = None  # MANDATORY: Subclasses must override this (tuple of str)
    HAS_NAME = False  # OPTIONAL: Subclasses may override this to denote the entity already has a name field,
                      #           in which case that name field will be renamed to "_name"

    @classmethod
    def get_all(cls, api, *args, **kwargs):
        pluck = api.create_pluck(cls.CONTAINER_SUBCOMPONENTS_FIELD)
        # Components do not support filtering, so filter on client side
        container_id = kwargs.get("container_id")
        if container_id:
            pluck = pluck.pluck(cls.CONTAINER_TYPE.ID_FIELD)
        containers = cls.CONTAINER_TYPE.iter_all(api, pluck=pluck)
        if container_id is not None:
            containers = (c for c in containers if c[cls.CONTAINER_TYPE.ID_FIELD] == container_id)
        entities = [cls._entity_hook(entity)
                    for container in containers
                    for entity in container[cls.CONTAINER_SUBCOMPONENTS_FIELD]
                    if cls._entity_predicate(entity)]
        if kwargs.get("hook"):
            entities = kwargs["hook"](entities)
        return entities
    @classmethod
    def format_name(cls, *parts):
        assert len(parts) == len(cls.ENTITY_NAME_PART_FIELDS), \
               "Expected {} parts, got {}".format(len(cls.ENTITY_NAME_PART_FIELDS), len(parts))
        return cls.ENTITY_NAME_FORMATTER.format(*parts)
    @classmethod
    def _entity_hook(cls, entity):
        if cls.HAS_NAME:
            entity["_"+cls.NAME_FIELD] = entity[cls.NAME_FIELD]
        entity[cls.NAME_FIELD] = cls.format_name(*(entity[field] for field in cls.ENTITY_NAME_PART_FIELDS))
        return entity
    @classmethod
    def _entity_predicate(cls, entity):
        return True
    @classmethod
    def iter_all(cls, api, *args, **kwargs):
        return iter(cls.get_all(api, *args, **kwargs))
    @classmethod
    def get_by_name(cls, api, name):
        match = cls.NAME_PATTERN.match(name)
        if not match:
            raise NoSuchEntity('"{}" is not a valid {} name'.format(name, cls.get_entity_name()))
        ids = (long(id) for id in match.groups())
        try:
            return cls._get_entity_by_ids(api, *ids)
        except ManagementException as e:
            if e.http_code != 404:
                raise
            raise NoSuchEntity('No such {}: "{}"'.format(cls.get_entity_name(), name))
    @classmethod
    def _get_entity_by_ids(cls, api, container_id, entity_id):
        entity = api.get(cls.CONTAINER_TYPE.get_uri(container_id, cls.CONTAINER_SUBCOMPONENTS_FIELD, entity_id))
        entity = cls._entity_hook(entity)
        cls._validate_fetched_entity(entity)
        return entity
    @classmethod
    def get_component_uri(cls, component, *parts):
        return cls.CONTAINER_TYPE.get_uri(
            component[cls.SUBCOMPONENT_CONTAINER_ID_FIELD],
            cls.CONTAINER_SUBCOMPONENTS_FIELD,
            component[cls.ID_FIELD],
            *parts)



################################################################################
#  HIERARCHICAL TYPE BASE CLASSES                                              #
################################################################################

def create_hierarchical_type(
    super_type,
    uri,
    master_class_name,
    master_mgmt_name,
    master_description_subject,
    search_type_field,
    ):
    super_type_name = master_class_name + "SuperType"
    class SuperType(super_type, TypeChoicesHelpMixin):
        class __metaclass__(getattr(super_type, '__metaclass__', type)):
            def __new__(mcls, name, bases, attrs):
                cls = getattr(super_type, '__metaclass__', type).__new__(mcls, name, bases, attrs)
                # Map subtype names to subtype entity classes
                if name != super_type_name and cls.ENTITY_MGMT_SUBTYPE:
                    SuperType._types[cls.ENTITY_MGMT_SUBTYPE] = cls
                    cls.SuperType = SuperType
                return cls
        TYPE_NAME = None  # Optionally override this on subclasses
        URI = uri  # Do not override this
        ENTITY_MGMT_SUBTYPE = None  # MANDATORY: Override this
        ENTITY_MGMT_NAME = master_mgmt_name  # Do not override this
        SEARCH_TYPE_FIELD = search_type_field  # Do not override this
        DESCRIPTION = None  # Do not override this
        DESCRIPTION_SUBJECT = master_description_subject  # MANDATORY: Override this
        QUERY_SORT = QuerySort().asc("name")
        _types = OrderedDict()
        @classmethod
        def get_name(cls):
            return combine_type_names(*cls._types.itervalues())
        def get_type_name_parts(self):
            return [t.get_name() for t in self._types.itervalues()]
        @classmethod
        def get_entity_name(cls):
            return cls.DESCRIPTION_SUBJECT
        @classmethod
        def mgmt_subtype_to_class(cls, type_name):
            entity_type = cls._types.get(type_name.upper())
            if entity_type is None:
                raise AssertionError("Type not registered: " + type_name)
            return entity_type
        @classmethod
        def get_description(cls):
            if not hasattr(cls, "__description"):
                cls.__description = "Name of an existing {}".format(cls.DESCRIPTION_SUBJECT)
            return cls.__description

    SuperType.__name__ = super_type_name

    class SubtypeMixin(object):
        TYPE_NAME = None
        @classmethod
        def get_name(cls):
            # This is not a mistake - the intention is to use super(SuperType)
            # and not super(SubtypeMixin).
            return super(SuperType, cls).get_name()
        def get_type_name_parts(self):
            return [self.get_name()]
        @classmethod
        def _fetch_entity_by_name(cls, api, name):
            # Call API explicitly instead of get_all() or iter_all(),
            # so that _get_query_params() will not be used
            entities = api.paged_query(cls.get_uri(), params=dict(name=name))
            if entities is not None:
                for entity in entities:
                    if entity.name == name:
                        return entity
            return None

    SubtypeMixin.__name__ = master_class_name + "SubtypeMixin"
    SuperType.SubtypeMixin = SubtypeMixin

    class SubtypeBase(SubtypeMixin, SuperType):
        class __metaclass__(SuperType.__metaclass__):
            def __new__(mcls, name, bases, attrs):
                if name != "SubtypeBase":
                    is_concrete_subtype = any(map((lambda base: hasattr(base, "__concrete_hierarchical_subtype__")), bases))
                    if not is_concrete_subtype:
                        attrs["__concrete_hierarchical_subtype__"] = True
                        mgmt_subtype = attrs.get("ENTITY_MGMT_SUBTYPE")
                        assert mgmt_subtype, name + " must declare ENTITY_MGMT_SUBTYPE"
                        attrs["QUERY_FILTER"] = QueryFilter(type=mgmt_subtype.lower())
                return SuperType.__metaclass__.__new__(mcls, name, bases, attrs)
        @classmethod
        def _validate_fetched_entity(cls, entity):
            if entity.type.lower() != cls.ENTITY_MGMT_SUBTYPE.lower():
                raise NoSuchEntity('No such {expected_type}: "{name}" is a {actual_type}, not a {expected_type}'.format(
                    name=entity.name,
                    actual_type=cls.mgmt_subtype_to_class(entity.type).get_entity_name(),
                    expected_type=cls.get_entity_name(),
                    ))
        @classmethod
        def _search_hook(cls, entities):
            return [e for e in entities
                    if e and (e.properties.get(cls.SEARCH_TYPE_FIELD) or "").lower() == cls.ENTITY_MGMT_SUBTYPE.lower()]

    SubtypeBase.__name__ = master_class_name + "SubtypeBase"

    return SuperType, SubtypeMixin, SubtypeBase

def create_combined_hierarchical_subtype(
    types,
    class_name,
    entity_name,
    entity_description_subject,
    ):
    assert len(types) >= 2, "At least two types must be specified"
    assert len(set(t.SuperType for t in types)) == 1, "Combined types must share a common super-type"
    SuperType = types[0].SuperType
    class CombinedSubtype(SuperType.SubtypeMixin, SuperType):
        TYPE_NAME = combine_type_names(*types)
        ENTITY_MGMT_SUBTYPE = None
        ENTITY_MGMT_SUBTYPES = set(t.ENTITY_MGMT_SUBTYPE.lower() for t in types)
        ENTITY_SUBTYPES = set(t.ENTITY_MGMT_SUBTYPE.lower() for t in types)
        ENTITY_NAME = entity_name
        DESCRIPTION_SUBJECT = entity_description_subject
        def get_type_name_parts(self):
            return [t.get_name() for t in types]
        @classmethod
        def _is_correct_subtype(cls, subtype):
            return (subtype or "").lower() in cls.ENTITY_SUBTYPES
        @classmethod
        def _validate_fetched_entity(cls, entity):
            if not cls._is_correct_subtype(entity.type):
                raise NoSuchEntity('No such {expected_type}: "{name}" is a {actual_type}, not a {expected_type}'.format(
                    name=entity.name,
                    actual_type=cls.mgmt_subtype_to_class(entity.type).get_entity_name(),
                    expected_type=cls.DESCRIPTION_SUBJECT,
                    ))
        @classmethod
        def _query_hook(cls, results):
            return [e for e in results if cls._is_correct_subtype(e.type)]
        @classmethod
        def _search_hook(cls, entities):
            return [e for e in entities if e and cls._is_correct_subtype(e.properties.get(cls.SEARCH_TYPE_FIELD))]

    CombinedSubtype.__name__ = class_name
    return CombinedSubtype



################################################################################
#  STORAGE DATA ENTITIES                                                       #
################################################################################

class DatasetBase(Entity):
    pass

Dataset, DatasetSubtypeMixin, DatasetSubtypeBase = create_hierarchical_type(
    super_type=DatasetBase,
    uri="datasets",
    master_class_name="Dataset",
    master_mgmt_name="VOLUME",
    master_description_subject="dataset, snapshot or clone",
    search_type_field="dataset_type",
    )

Dataset.SEARCHABLE = False
DatasetSubtypeBase.SEARCHABLE = False
Dataset.HOOK_DEPENDENCIES = ["type", "dataset_type"]

class MasterDataset(DatasetSubtypeBase):
    TYPE_NAME = "DATASET"
    ENTITY_NAME = "dataset"
    ENTITY_SHORT_NAME = "vol"
    ENTITY_MGMT_SUBTYPE = "MASTER"
    DESCRIPTION_SUBJECT = "dataset"

class DatasetSnap(DatasetSubtypeBase):
    TYPE_NAME = "SNAP"
    ENTITY_NAME = "snapshot"
    ENTITY_SHORT_NAME = "snap"
    ENTITY_MGMT_SUBTYPE = "SNAP"
    DESCRIPTION_SUBJECT = "snapshot"

class DatasetClone(DatasetSubtypeBase):
    TYPE_NAME = "CLONE"
    ENTITY_NAME = "clone"
    ENTITY_SHORT_NAME = "clone"
    ENTITY_MGMT_SUBTYPE = "CLONE"
    DESCRIPTION_SUBJECT = "clone"

DatasetOrClone = create_combined_hierarchical_subtype(
    types=[MasterDataset, DatasetClone],
    class_name="DatasetOrClone",
    entity_name="dataset",
    entity_description_subject="dataset or clone",
    )

DatasetSnapOrClone = create_combined_hierarchical_subtype(
    types=[DatasetSnap, DatasetClone],
    class_name="DatasetSnapOrClone",
    entity_name="dataset",
    entity_description_subject="snapshot or clone",
    )

Volume, VolumeSubtypeMixin, VolumeSubtypeBase = create_hierarchical_type(
    super_type=DatasetBase,
    uri="volumes",
    master_class_name="Volume",
    master_mgmt_name="VOLUME",
    master_description_subject="volume, snapshot or clone",
    search_type_field="volume_type",
    )

class MasterVolume(VolumeSubtypeBase):
    TYPE_NAME = "VOL"
    ENTITY_NAME = "volume"
    ENTITY_SHORT_NAME = "vol"
    ENTITY_MGMT_SUBTYPE = "MASTER"
    DESCRIPTION_SUBJECT = "volume"

class VolumeSnap(VolumeSubtypeBase):
    TYPE_NAME = "SNAP"
    ENTITY_NAME = "volume snapshot"
    ENTITY_SHORT_NAME = "vol-snap"
    ENTITY_MGMT_SUBTYPE = "SNAP"
    DESCRIPTION_SUBJECT = "volume snapshot"

class VolumeClone(VolumeSubtypeBase):
    TYPE_NAME = "CLONE"
    ENTITY_NAME = "volume clone"
    ENTITY_SHORT_NAME = "vol-clone"
    ENTITY_MGMT_SUBTYPE = "CLONE"
    DESCRIPTION_SUBJECT = "volume clone"

VolumeOrClone = create_combined_hierarchical_subtype(
    types=[MasterVolume, VolumeClone],
    class_name="VolumeOrClone",
    entity_name="volume",
    entity_description_subject="volume or clone",
    )

VolumeSnapOrClone = create_combined_hierarchical_subtype(
    types=[VolumeSnap, VolumeClone],
    class_name="VolumeSnapOrClone",
    entity_name="volume",
    entity_description_subject="snapshot or clone",
    )

Filesystem, FilesystemSubtypeMixin, FilesystemSubtypeBase = create_hierarchical_type(
    super_type=DatasetBase,
    uri="filesystems",
    master_class_name="Filesystem",
    master_mgmt_name="FILESYSTEM",
    master_description_subject="filesystem, snapshot or clone",
    search_type_field="filesystem_type",
    )

class MasterFilesystem(FilesystemSubtypeBase):
    TYPE_NAME = "FS"
    ENTITY_NAME = "filesystem"
    ENTITY_SHORT_NAME = "fs"
    ENTITY_MGMT_SUBTYPE = "MASTER"
    DESCRIPTION_SUBJECT = "filesystem"

class FilesystemSnap(FilesystemSubtypeBase):
    TYPE_NAME = "SNAP"
    ENTITY_NAME = "filesystem snapshot"
    ENTITY_SHORT_NAME = "fs-snap"
    ENTITY_MGMT_SUBTYPE = "SNAP"
    DESCRIPTION_SUBJECT = "filesystem snapshot"

class FilesystemClone(FilesystemSubtypeBase):
    TYPE_NAME = "CLONE"
    ENTITY_NAME = "filesystem clone"
    ENTITY_SHORT_NAME = "fs-clone"
    ENTITY_MGMT_SUBTYPE = "CLONE"
    DESCRIPTION_SUBJECT = "filesystem clone"

FilesystemOrSnap = create_combined_hierarchical_subtype(
    types=[MasterFilesystem, FilesystemSnap],
    class_name="FilesystemOrSnap",
    entity_name="filesystem",
    entity_description_subject="filesystem or snap",
    )

FilesystemOrClone = create_combined_hierarchical_subtype(
    types=[MasterFilesystem, FilesystemClone],
    class_name="FilesystemOrClone",
    entity_name="filesystem",
    entity_description_subject="filesystem or clone",
    )

FilesystemSnapOrClone = create_combined_hierarchical_subtype(
    types=[FilesystemSnap, FilesystemClone],
    class_name="FilesystemSnapOrClone",
    entity_name="filesystem",
    entity_description_subject="snapshot or clone",
    )

ConsistencyOrSnapGroup, ConsistencyGroupSubtypeMixin, ConsistencyGroupSubtypeBase = create_hierarchical_type(
    super_type=DatasetBase,
    uri="cgs",
    master_class_name="ConsistencyOrSnapGroup",
    master_mgmt_name="CG",
    master_description_subject="consistency group or snapshot group",
    search_type_field="cg_type",
    )

ConsistencyOrSnapGroup.SEARCHABLE = False  # Workaround for INFINIBOX-15072
ConsistencyGroupSubtypeBase.SEARCHABLE = False  # Workaround for INFINIBOX-15072

class ConsistencyGroup(ConsistencyGroupSubtypeBase):
    TYPE_NAME = "CG"
    ENTITY_NAME = "consistency group"
    ENTITY_SHORT_NAME = "cg"
    ENTITY_MGMT_SUBTYPE = "MASTER"
    DESCRIPTION_SUBJECT = "consistency group"

class SnapGroup(ConsistencyGroupSubtypeBase):
    TYPE_NAME = "SG"
    ENTITY_NAME = "snapshot group"
    ENTITY_SHORT_NAME = "sg"
    ENTITY_MGMT_SUBTYPE = "SNAP"
    DESCRIPTION_SUBJECT = "snapshot group"

class Pool(Entity):
    QUERY_SORT = QuerySort().asc("name")
    pass

class Host(Entity):
    QUERY_SORT = QuerySort().asc("name")
    pass

class Cluster(Entity):
    QUERY_SORT = QuerySort().asc("name")
    pass

class ReplicatedEntityBase(TypeChoicesHelpMixin, Entity):
    class __metaclass__(type):
        def __new__(mcls, name, bases, attrs):
            if name != "ReplicatedEntityBase":
                attrs["_ENTITY_TYPES"] = tuple(mcls.inherit_entity_cls(entity_cls, attrs["_FOR_REPLICATED_ENTITIES"])
                                               for entity_cls in attrs.get("_ENTITY_TYPES", ReplicatedEntityBase._ENTITY_TYPES))
            return type.__new__(mcls, name, bases, attrs)
        @classmethod
        def inherit_entity_cls(mcls, entity_cls, for_replicated_entities):
            class subclass(entity_cls):
                _REPLICATED = for_replicated_entities
            subclass.__name__ = "_Replicated" + entity_cls.__name__
            return subclass

    class _Volume(Entity):
        URI = "volumes"
        TYPE_NAME = "VOL"
        SEARCHABLE = False
        QUERY_FILTER = QueryFilter().eq("type", "MASTER").eq("cg_id", "null")
        HOOK_DEPENDENCIES = ["cg_id", "rmr_source", "rmr_target"]
        _REPLICATED = None  # See ReplicatedEntityBase.__metaclass__
        @classmethod
        def _query_hook(cls, results):
            for entity in results:
                if (entity.rmr_source or entity.rmr_target) == cls._REPLICATED:
                    yield entity
        @classmethod
        def _fetch_entity_by_name(cls, *args, **kwargs):
            entity = super(ReplicatedEntityBase._Volume, cls)._fetch_entity_by_name(*args, **kwargs)
            if entity:
                entity._is_dataset = True
            return entity

    class _Cg(Entity):
        URI = "cgs"
        TYPE_NAME = "CG"
        SEARCHABLE = False
        QUERY_FILTER = QueryFilter().eq("type", "MASTER")
        HOOK_DEPENDENCIES = ["is_replicated"]
        _REPLICATED = None  # See ReplicatedEntityBase.__metaclass__
        @classmethod
        def _query_hook(cls, results):
            for entity in results:
                if entity.is_replicated == cls._REPLICATED:
                    yield entity
        @classmethod
        def _fetch_entity_by_name(cls, *args, **kwargs):
            entity = super(ReplicatedEntityBase._Cg, cls)._fetch_entity_by_name(*args, **kwargs)
            if entity:
                entity._is_dataset = False
            return entity

    _ENTITY_TYPES = (_Cg, _Volume)  # See __metaclass__
    _FOR_REPLICATED_ENTITIES = None  # See __metaclass__
    def get_type_name_parts(self):
        return [t.get_name() for t in self._ENTITY_TYPES]
    @classmethod
    def paged_query(cls, *args, **kwargs):
        raise NotImplementedError("{} does not support paged_query()".format(cls.__name__))
    @classmethod
    def query(cls, *args, **kwargs):
        raise NotImplementedError("{} does not support query()".format(cls.__name__))
    @classmethod
    def _fetch_entity_by_name(cls, *args, **kwargs):
        for entity_cls in cls._ENTITY_TYPES:
            entity = entity_cls._fetch_entity_by_name(*args, **kwargs)
            if entity:
                return entity
        return None
    def get_possible_values(self, *args, **kwargs):
        for entity_cls in self._ENTITY_TYPES:
            for value in entity_cls().get_possible_values(*args, **kwargs):
                yield value

class ReplicatedEntity(ReplicatedEntityBase):
    _FOR_REPLICATED_ENTITIES = True
    ENTITY_NAME = "volume or consistency group with replication"

class NonReplicatedEntity(ReplicatedEntityBase):
    _FOR_REPLICATED_ENTITIES = False
    ENTITY_NAME = "volume or consistency group without replication"

class ReplicatedVolume(ReplicatedEntityBase):
    _FOR_REPLICATED_ENTITIES = True
    _ENTITY_TYPES = (ReplicatedEntityBase._Volume,)
    ENTITY_NAME = "volume with replication"

class NonReplicatedVolume(ReplicatedEntityBase):
    _FOR_REPLICATED_ENTITIES = False
    _ENTITY_TYPES = (ReplicatedEntityBase._Volume,)
    ENTITY_NAME = "volume without replication"

class ReplicatedCg(ReplicatedEntityBase):
    _FOR_REPLICATED_ENTITIES = True
    _ENTITY_TYPES = (ReplicatedEntityBase._Cg,)
    ENTITY_NAME = "consistency group with replication"

class NonReplicatedCg(ReplicatedEntityBase):
    _FOR_REPLICATED_ENTITIES = False
    _ENTITY_TYPES = (ReplicatedEntityBase._Cg,)
    ENTITY_NAME = "consistency group without replication"



################################################################################
#  NAS                                                                         #
################################################################################

class Export(Entity):
    SEARCHABLE = False
    NAME_FIELD = "export_path"
    QUERY_SORT = QuerySort().asc("export_path")

class ExportClient(RegexType):
    # Uncomment these members once hostnames and wildcards are supported:
    # IP_SEGMENT_OR_WILDCARD_PATTERN = r"(25[0-5]|2[0-4][0-9]|[01]?[0-9][0-9]?)|\*"
    # IP_WITH_WILDCARD_PATTERN = r"({0}\.){{3}}{0}".format(IP_SEGMENT_OR_WILDCARD_PATTERN)
    # HOSTNAME_WITH_WILDCARD_PATTERN = r"[a-zA-Z0-9\-.*]"  # RFC-1123, plus wildcards
    # PATTERN = r"^(?P<hostname>{0})|((?P<start_ip>{1})(-(?P<end_ip>{1})))$" \
    #           .format(HOSTNAME_WITH_WILDCARD_PATTERN, IP_WITH_WILDCARD_PATTERN)

    # Delete these members once hostnames and wildcards are supported:
    IP_SEGMENT_PATTERN = r"(25[0-5]|2[0-4][0-9]|[01]?[0-9][0-9]?)"
    IP_PATTERN = r"({0}\.){{3}}{0}".format(IP_SEGMENT_PATTERN)
    PATTERN = r"({0}(-{0})?)|\*".format(IP_PATTERN)

    def _get_pattern(self):
        return self.PATTERN
    def _get_error(self):
        return "is not a valid client identifier"
    @classmethod
    def get_name(cls):
        return "CLIENT"
    # Uncomment this description once hostnames and wildcards are supported:
    # def get_description(self):
    #     return 'A hostname, IP address, IP range (e.g 10.0.0.1-10.0.0.10), or ' \
    #            'IP with wildcard (e.g 10.0.0.*)'

    # Delete this description once hostnames and wildcards are supported:
    def get_description(self):
        return 'A full wildcard (i.e: *), IP address or IP range (e.g 10.0.0.1-10.0.0.10)'

class ExportAccessLevel(ChoiceType):
    LEVELS = NamedOrderedDict.constants("RO", "RW")
    MGMT_LEVELS = Munch((v, k) for k, v in LEVELS.iteritems())
    def _get_values(self):
        return self.LEVELS.iterkeys()

# class ExportPermissions(RegexType):
#     PERMISSIONS = NamedOrderedDict.constants("ro", "rw")
#     PERM_TYPE_NAMES = PERMISSIONS.values()
#     MATCH_PATTERN = r"^\s*(({})\(.*?\)\s*)+$".format("|".join(PERM_TYPE_NAMES))
#     COMPILED_SEARCH_PATTERN = re.compile(r"\s*(?P<perm>{})\((?P<hostspec>[^)]+)\)\s*".format("|".join(PERM_TYPE_NAMES)))
#     def _get_pattern(self):
#         return self.MATCH_PATTERN
#     def _get_flags(self):
#         return re.IGNORECASE
#     def _get_error(self):
#         return "is not a valid export permissions string"
#     @classmethod
#     def get_name(cls):
#         return "ACL"
#     def get_description(self):
#         return 'NFS export permissions string, consisting of multiple "perm(host[,host,...])" values, such that "perm" ' \
#                'may be one of {}, and "host" may contain a hostname, IP address, IP range (e.g 10.0.0.1-10.0.0.10), ' \
#                'or IP with wildcard (e.g 10.0.0.*)'.format(", ".join('"{}"'.format(p) for p in self.PERM_TYPE_NAMES))
#     def _get_actual(self, value):
#         permissions = {perm : "" for perm in self.PERM_TYPE_NAMES}
#         for match in self.COMPILED_SEARCH_PATTERN.finditer(value):
#             permissions[match.group("perm")] += ",{}".format(match.group("hostspec"))
#         permissions = self.format_acl(permissions)
#         return True, permissions
#     def get_possible_values(self, api, starts_with=None):
#         if not starts_with:
#             return (p + "(" for p in self.PERM_TYPE_NAMES)
#         partial_value_start_index = 0
#         matches = list(self.COMPILED_SEARCH_PATTERN.finditer(starts_with))
#         partial_value_start_index = matches[-1].end() if matches else 0
#         complete_part, partial_part = starts_with[:partial_value_start_index], starts_with[partial_value_start_index:]
#         if not partial_part.strip():
#             complete_part += partial_part
#             partial_part = ""
#         return ["{}{}(".format(complete_part, name) for name in self.PERM_TYPE_NAMES if name.starts_with(partial_part)]
#     @classmethod
#     def format_acl(cls, acl_dict):
#         for perm in cls.PERMISSIONS.iterkeys():
#             if acl_dict.get(cls.PERMISSIONS[perm],False):
#                 acl_dict[perm] = acl_dict.pop(cls.PERMISSIONS[perm])
#         return acl_dict

################################################################################
#  HARDWARE                                                                    #
################################################################################

class Node(Entity):
    URI = "components/nodes"
    SEARCHABLE = False
    PAGED = False
    def get_description(self):
        return "Node name"

class PatchPanel(Entity):
    URI = "components/patch_panels"
    SEARCHABLE = False
    PAGED = False
    def get_description(self):
        return "Patch panel name"

class Initiator(Entity):
    URI = "initiators"
    SEARCHABLE = False
    FILTERABLE = False
    PAGED = True
    def get_description(self):
        return "WWPN address"

class LoggedInInitiator(Initiator):
    NAME_FIELD = "address"
    HOOK_DEPENDENCIES = ["address", "host_id"]
    PATTERN = r"^([0-9a-fA-F]{2}:?){7}[0-9a-fA-F]{2}$"
    def _verify_raw(self, value):
        pattern = re.compile(self.PATTERN)
        return "is not a valid address" if not pattern.match(value) else None
    @classmethod
    def _query_hook(self, results):
        return [i for i in results if i.host_id == 0]
    @classmethod
    def _get_dynamic(self, api, arg, value):
        return arg, value
    def get_description(self):
        return "Logged in initiator"

class FCPort(ComponentNestedEntity):
    TYPE_NAME = "PORT"
    ENTITY_NAME = "FC port"
    CONTAINER_TYPE = Node
    CONTAINER_SUBCOMPONENTS_FIELD = "fc_ports"
    SUBCOMPONENT_CONTAINER_ID_FIELD = "node_index"
    NAME_PATTERN = re.compile(r"^N(\d+)FC(\d+)$", re.I)
    ENTITY_NAME_FORMATTER = "N{}FC{}"
    ENTITY_NAME_PART_FIELDS = "node_index", "id"
    _WWN_PATTERN = re.compile(WWN.PATTERN)
    @classmethod
    def format_name(cls, *parts):
        if parts.count(None) != 0:
            return None
        return super(FCPort, cls).format_name(*parts)
    @classmethod
    def get_by_name(cls, api, name):
        if cls._WWN_PATTERN.match(name):
            for port in cls.iter_all(api):
                if port.wwpn == name:
                    return cls._entity_hook(port)
        return super(FCPort, cls).get_by_name(api, name)
    def get_description(self):
        return "Node and FC port index, e.g: N2FC5, or WWPN (a 16-digit hexadecimal " \
               "number, optionally divided by colons to 8 groups of two digits each)"

class EthPort(ComponentNestedEntity):
    TYPE_NAME = "PORT"
    ENTITY_NAME = "ethernet port"
    CONTAINER_TYPE = Node
    CONTAINER_SUBCOMPONENTS_FIELD = "eth_ports"
    SUBCOMPONENT_CONTAINER_ID_FIELD = "node_index"
    NAME_PATTERN = re.compile(r"^N(\d+)ETH(\d+)$", re.I)
    ENTITY_NAME_FORMATTER = "N{}ETH{}"
    ENTITY_NAME_PART_FIELDS = "node_index", "port_number"
    HAS_NAME = True
    _VISIBLE_PORT_ROLES = {"DATA", "MANAGEMENT"}
    @classmethod
    def _entity_predicate(cls, entity):
        return not system.user.roles.is_customer() or entity.get("role", "").upper() in cls._VISIBLE_PORT_ROLES
    def get_description(self):
        return "Node and ethernet port index, e.g: N1ETH2"

class UsableEthPort(EthPort):
    @classmethod
    def _validate_fetched_entity(cls, entity):
        if not cls._entity_predicate(entity):
            raise NoSuchEntity('{} "{}" cannot be used for this command, as its role is "{}"'
                               .format(cls.get_entity_name(capitalize=True), entity.name, entity.role))
    def get_colliding_values(self, api,existing_values, possible_values, current_val):
        if existing_values:
            return [v for v in possible_values if v[1] != next(iter(existing_values))[1]]
        return []

    @classmethod
    def _entity_predicate(cls, entity):
        return (entity.get("role") or "").upper() in ("DATA", "MANAGEMENT")

class PatchPanelFrame(ComponentNestedEntity):
    TYPE_NAME = "PATCHPANELPORT"
    ENTITY_NAME = "patch panel port"
    CONTAINER_TYPE = PatchPanel
    CONTAINER_SUBCOMPONENTS_FIELD = "frames"
    SUBCOMPONENT_CONTAINER_ID_FIELD = 1  # No such field, but there's only one patch panel
    NAME_PATTERN = re.compile(r"^PP(\d+)F(\d+)$", re.I)
    ENTITY_NAME_FORMATTER = "PP{}F{}"
    ENTITY_NAME_PART_FIELDS = None, "location"
    @classmethod
    def _entity_hook(cls, entity):
        entity.name = "PP1F{}".format(entity.location)
        return entity

class NodeDrive(ComponentNestedEntity):
    ENTITY_NAME = "node drive"
    CONTAINER_TYPE = Node
    CONTAINER_SUBCOMPONENTS_FIELD = "drives"
    SUBCOMPONENT_CONTAINER_ID_FIELD = "node_index"
    NAME_PATTERN = re.compile(r"^N(\d+)D(\d+)$", re.I)
    ENTITY_NAME_FORMATTER = "N{}D{:02}"
    ENTITY_NAME_PART_FIELDS = "node_index", "drive_index"
    def get_description(self):
        return "Node and drive index, e.g: N1D8"

class NodeHdd(NodeDrive):
    @classmethod
    def _entity_predicate(cls, entity):
        return entity.type != "SSD"

class NodeSsd(NodeDrive):
    @classmethod
    def _entity_predicate(cls, entity):
        return entity.type == "SSD"
    NAME_PATTERN = re.compile(r"^N(\d+)SSD(\d+)$", re.I)
    ENTITY_NAME_FORMATTER = "N{}SSD{:02}"
    def get_description(self):
        return "Node and SSD index, e.g: N1D8"

class Enclosure(Entity):
    SEARCHABLE = False
    PAGED = False
    URI = "components/enclosures"
    HOOK_DEPENDENCIES = ["id"]
    @classmethod
    def _query_hook(cls, enclosures):
        for e in enclosures:
            e.name = "E{}".format(e.id)
        return enclosures
    def get_description(self):
        return "Enclosure name"

class Drive(ComponentNestedEntity):
    ENTITY_NAME = "data drive"
    CONTAINER_TYPE = Enclosure
    CONTAINER_SUBCOMPONENTS_FIELD = "drives"
    SUBCOMPONENT_CONTAINER_ID_FIELD = "enclosure_index"
    NAME_PATTERN = re.compile(r"^E(\d+)D(\d+)$", re.I)
    ENTITY_NAME_FORMATTER = "E{}D{:02}"
    ENTITY_NAME_PART_FIELDS = "enclosure_index", "drive_index"
    def get_description(self):
        return "Enclosure and drive index, e.g: E2D17"



################################################################################
#  EVENTS AND NOTIFICATIONS                                                    #
################################################################################

class Event(Entity):
    SEARCHABLE = False
    QUERY_SORT = QuerySort().asc("timestamp")
    @classmethod
    def get_by_seq(cls, api, seq):
        return cls.get_single(api, filter=api.create_filter(seq_num=seq))

class EventRule(Entity):
    URI = "notifications/rules"
    SEARCHABLE = False
    QUERY_SORT = QuerySort().asc("name")

class EventInfo(Entity):
    URI = "events/types"
    ID_FIELD = "code"
    NAME_FIELD = "code"
    CACHEABLE = True
    SEARCHABLE = False
    FILTERABLE = False
    PLUCKABLE = False
    PAGED = False
    def get_description(self):
        "Event code"
    @classmethod
    def _query_hook(cls, events_info):
        return sorted(events_info.codes, key=lambda event_info: event_info.code)

class EventLevel(ConstantEntity):
    TYPE_NAME = "LEVEL"  # For subclasses
    ENTITY_NAME = "event level"  # For subclasses
    URI = "events/types"
    ATTR = "levels"
    def get_description(self):
        return "Event severity level"
    @classmethod
    def _transform_values(cls, levels):
        cls._value_by_level = { level.name : level.value for level in levels }
        return (level.name for level in sorted(levels, key=lambda level: cls._value_by_level[level.name]))
    @classmethod
    def sort(cls, api, levels):
        cls._ensure_values(api)
        return sorted(levels, key=lambda level: cls._value_by_level[level])

class EventLevelOrNone(ConstantsHelpMixin, EventLevel):
    CONSTANTS = Munch(NONE=NONE_CONSTANT)

class EventCode(ConstantEntity):
    TYPE_NAME = "CODE"  # For subclasses
    ENTITY_NAME = "event code"  # For subclasses
    URI = "events/types"
    ATTR = "codes"
    FIELD = "code"
    def get_description(self):
        return "Event code"

class EventCodeOrNone(ConstantsHelpMixin, EventCode):
    CONSTANTS = Munch(NONE=NONE_CONSTANT)

class EventVisibility(ConstantEntity):
    TYPE_NAME = "VISIBILITY"
    URI = "events/types"
    ATTR = "visibilities"
    def get_description(self):
        return "Event visibility scope"

class EventReporter(ConstantEntity):
    TYPE_NAME = "REPORTER"
    URI = "events/types"
    ATTR = "reporters"
    def get_description(self):
        return "Event reporter"

class NotificationTarget(Entity):
    URI = "notifications/targets"
    SEARCHABLE = False
    QUERY_FILTER = SubtypeQueryFilter("protocol", "PROTOCOL")
    QUERY_SORT = QuerySort().asc("name")
    PROTOCOL = None  # MANDATORY: Subclasses must override this

class SmtpTarget(NotificationTarget):
    PROTOCOL = "SMTP"
    ENTITY_NAME = "SMTP notification target"

class SnmpTarget(NotificationTarget):
    PROTOCOL = "SNMP"
    ENTITY_NAME = "SNMP notification target"

class SnmpProtocol(MappedChoiceType):
    SNMPv1, SNMPv2c, SNMPv3 = ["SNMPv1", "SNMPv2c", "SNMPv3"]
    def _get_mapping(self):
        return (
            ("1",  self.SNMPv1),
            ("2c", self.SNMPv2c),
            ("3",  self.SNMPv3),
            )

class SnmpSecurity(ChoiceType):
    _VALUES = [NoAuthNoPriv, AuthNoPriv, AuthPriv] = ["NoAuthNoPriv", "AuthNoPriv", "AuthPriv"]
    def _get_values(self):
        return self._VALUES

class SnmpEncryption(ChoiceType):
    def _get_values(self):
        return ["AES", "DES"]

class SnmpAuthProtocol(ChoiceType):
    def _get_values(self):
        return ["MD5", "SHA"]

class SyslogTarget(NotificationTarget):
    PROTOCOL = "SYSLOG"
    ENTITY_NAME = "Syslog notification target"

class SyslogProtocol(ChoiceType):
    def _get_values(self):
        return "TCP", "UDP"

class SyslogFacility(ChoiceType):
    def _get_values(self):
        return ["local{}".format(n) for n in xrange(8)]



################################################################################
#  USER MANAGEMENT                                                             #
################################################################################

class User(Entity):
    URI = "users"
    SEARCHABLE = False
    QUERY_SORT = QuerySort().asc("name")
    QUERY_FILTER = SubtypeQueryFilter("type", "USER_TYPE")
    USER_TYPE = None  # MANDATORY: Subclasses must override this

class LocalUser(User):
    USER_TYPE = "Local"

class DisabledLocalUser(User):
    USER_TYPE = "Local"
    QUERY_FILTER = QueryFilter().eq("enabled", "false")
    def get_desciption():
        return "Name of an existing disabled local user"

class EnabledLocalUser(User):
    USER_TYPE = "Local"
    QUERY_FILTER = QueryFilter().eq("enabled", "true")
    def get_desciption():
        return "Name of an existing enabled local user"

class LdapGroup(User):
    USER_TYPE = "Ldap"

class PoolAdminMixin(object):
    HOOK_DEPENDENCIES = ["roles"]
    @classmethod
    def _query_hook(cls, users):
        from .utils import is_pool_admin  # Avoid cyclic imports
        return [u for u in users if is_pool_admin(u)]

class PoolAdmin(PoolAdminMixin, User):
    @classmethod
    def get_description(cls):
        return "Name of an existing {} or {}, with PoolAdmin role" \
               .format(LocalUser.get_entity_name(), LdapGroup.get_entity_name())

class PoolAdminUser(PoolAdminMixin, LocalUser):
    @classmethod
    def get_description(cls):
        return "Name of an existing {} with PoolAdmin role".format(LocalUser.get_entity_name())

class PoolAdminGroup(PoolAdminMixin, LdapGroup):
    @classmethod
    def get_description(cls):
        return "Name of an existing {} with PoolAdmin role".format(LdapGroup.get_entity_name())

class Username(ConstantsHelpMixin, LocalUser):
    TYPE_NAME = "USER"
    ENTITY_NAME = "USER"
    CONSTANTS = dict(
        INTERNAL="INTERNAL",
        )
    @classmethod
    def _fetch(cls, api, arg, value):
        return arg, value

class UserRole(ConstantEntity):
    TYPE_NAME = "ROLE"
    URI = "users/roles"
    def get_description(self):
        return "User role"

class UsableUserRole(UserRole):
    def get_possible_values(self, *args, **kwargs):
        values = super(UsableUserRole, self).get_possible_values(*args, **kwargs)
        if not values:
            return None
        if system.user.roles.is_infinidat():
            return values
        return (v for v in values if v.upper() in Roles.CUSTOMER_ROLES)

class MappableUserRole(UserRole):
    ENTITY_NAME = "mappable role"
    def get_possible_values(self, api, starts_with=None):
        values = super(MappableUserRole, self).get_possible_values(api, starts_with)
        if not values:
            return None
        return (v for v in values if v.upper() in Roles.CUSTOMER_ROLES)

class UsersRepository(Entity):
    ENTITY_NAME = "LDAP server"
    URI = "config/ldap"
    SEARCHABLE = False
    FILTERABLE = False
    QUERY_SORT = QuerySort().asc("name")
    QUERY_FILTER = SubtypeQueryFilter("repository_type", "REPOSITORY_TYPE")
    REPOSITORY_TYPE = None  # MANDATORY: Subclasses must override this

class Ldap(UsersRepository):
    ENTITY_NAME = "LDAP server"
    REPOSITORY_TYPE = "LDAP"

class ActiveDirectory(UsersRepository):
    ENTITY_NAME = "Active Directory server"
    REPOSITORY_TYPE = "ActiveDirectory"



################################################################################
#  NETWORK CONFIGURATION                                                       #
################################################################################

class NetworkInterface(Entity):
    TYPE_NAME = "INTERFACE"
    URI = "network/interfaces"
    SEARCHABLE = False
    QUERY_SORT = QuerySort().asc("name")

class NetworkInterfaceType(ChoiceType):
    def _get_values(self):
        return ("PORT_GROUP",)
    @classmethod
    def get_name(self):
        return "TYPE"

class NetworkSpace(Entity):
    URI = "network/spaces"
    SEARCHABLE = False
    QUERY_SORT = QuerySort().asc("name")

class NetworkSpaceService(MappedChoiceType):
    def _get_mapping(self):
        return {
            "NAS" : "NAS_SERVICE",
            "REPLICATION" : "RMR_SERVICE",
            }
    @classmethod
    def get_name(cls):
        return "SERVICE"

class BurstPriority(Range):
    MIN = 1
    MAX = 5
    def get_type_name_parts(self):
        return (str(self.MIN), str(self.MAX))
    @classmethod
    def get_name(self):
        return "Burst Priority"
    def get_description(self):
        return "Sets the priority of BW assignment to different Network Spaces " \
               "competing over the same interface. Lower value means higher priority."

class MetricsInterval(Range):
    MIN = 1
    MAX = 60
    def get_type_name_parts(self):
        return (str(self.MIN), str(self.MAX))
    @classmethod
    def get_name(self):
        return "Metrics Interval"
    def get_description(self):
        return "Metrics refresh interval. Max value is {}".format(self.MAX)

class NetworkRateLimit(Positive):
    @classmethod
    def get_name(self):
        return "LIMIT"
    def get_description(self):
        return "An amount of megabits per second"


class NetworkRateLimitOrNone(ConstantsHelpMixin, NetworkRateLimit):
    CONSTANTS = Munch(NONE=NONE_CONSTANT)

################################################################################
#  FC CONFIGURATION                                                            #
################################################################################

class FCSoftTargetCreationQuantity(Dynamic):
    MIN = 1
    DEFAULT_VALUE = 6
    def _get_dynamic(self, api, arg, value):
        max_q = api.get("config/limits/soft_targets.max_soft_targets_in_switch")
        if not isinstance(value, (int, str, basestring)):
            raise BadValue("Quantity must be between %s and %s: %s" % (self.MIN, max_q, value))
        if int(value) <= max_q and int(value) >= self.MIN:
            return arg, int(value)
        raise BadValue("Quantity must be between %s and %s: %s" % (self.MIN, max_q, value))


class FCPortRole(ChoiceType):
    HARD_PORT = "HARD_PORT"
    SOFT_PORT = "SOFT_PORT"
    def _get_values(self):
        return [self.HARD_PORT, self.SOFT_PORT]

class FCSwitch(Entity):
    TYPE_NAME = "SWITCH"
    URI = "fc/switches"
    SEARCHABLE = False
    QUERY_SORT = QuerySort().asc("name")

class FCSoftTarget(Entity):
    TYPE_NAME = "WWPN"
    URI = "fc/soft_targets"
    SEARCHABLE = False
    FILTERABLE = False
    CASE_INSENSITIVE_NAME = True
    NAME_FIELD = "wwpn"
    # QUERY_SORT = QuerySort().asc("node_id").asc("port_number")



################################################################################
#  REPLICATION                                                                 #
################################################################################

class ReplicationLink(Entity):
    TYPE_NAME = "LINK"
    URI = "links"
    SEARCHABLE = False

class ReplicationLinkBySystemName(ReplicationLink):
    TYPE_NAME = "SYSTEM"
    NAME_FIELD = "remote_system_name"
    @classmethod
    def get_entity_name(cls):
        return ReplicationLink.get_entity_name()

class Replica(Entity):
    SEARCHABLE = False
    NAME_FIELD = "id"

class ReplicaRole(ChoiceType):
    SOURCE = "SOURCE"
    TARGET = "TARGET"
    def _get_values(self):
        return [self.SOURCE, self.TARGET]
    @classmethod
    def get_name(cls):
        return "ROLE"

class ReplicaRoleOrNone(ReplicaRole):
    CONSTANTS = Munch(NONE=None)

class ReplicaDatasetPair(Name):
    _PAIR_SEPARATOR = ":"
    def _get_pattern(self):
        base_pattern = super(ReplicaDatasetPair, self)._get_pattern().lstrip(r"^").rstrip(r"$")
        return r"^{0}{1}{0}$".format(base_pattern, self._PAIR_SEPARATOR)
    def _get_actual(self, value):
        pair = value.split(self._PAIR_SEPARATOR, 1)
        assert len(pair) == 2
        return True, pair
    @classmethod
    def get_name(cls):
        return "LOCAL:REMOTE"
    def get_description(self):
        return "Pairs of local and remote dataset names, each delimited by a colon, " \
               "e.g: 'vol01:rep-vol01,vol02:rep-vol02'"



################################################################################
#  SERVICES                                                                    #
################################################################################

class SystemService(Entity):
    TYPE_NAME = "SERVICE"
    ENTITY_NAME = "service"
    PAGED = False
    FILTERABLE = False
    SEARCHABLE = False
    URI = "services"
    ID_FIELD = "name"
    HOOK_DEPENDENCIES = ["name"]
    @classmethod
    def _query_hook(cls, services):
        services.append(cls._get_mgmt_service())
        services.sort(key=lambda s: s.name)
        return services
    @classmethod
    def _get_mgmt_service(cls):
        mgmt_nodes = []
        for node in system.api.get("components/nodes?fields=id,services"):
            for service in node.services:
                if service.name == "mgmt":
                    mgmt_nodes.append(Munch(node_id=node.id,service=service))
                    break
        cluster_states = set([node.service.state for node in mgmt_nodes])
        return Munch(
            name="management",
            node_states=[Munch(
                node_id=node.node_id,
                state=node.service.state,
                role=node.service.role,
                ) for node in mgmt_nodes],
            cluster_state=cluster_states.pop() if len(cluster_states) == 1 else "DEGRADED",
            )


################################################################################
#  MISCELLANEOUS                                                               #
################################################################################

class Metadata(Entity):
    pass

class ConfigParameter(ConstantEntity):
    @classmethod
    def get_name(cls):
        return "PARAM"
    def get_description(self):
        return "Name of a configuration parameter"
    @classmethod
    def _request_values(cls, api):
        categories = api.get("config")
        names = []
        for category_name, category in categories.iteritems():
            names.extend(cls._get_param_names({ category_name : category }))
        names.sort()
        return names
    @classmethod
    def _get_param_names(cls, category):
        for name, value in category.iteritems():
            if isinstance(value, dict):
                for nested_name in cls._get_param_names(value):
                    yield "{}/{}".format(name, nested_name)
            else:
                yield name

class ConfigCategory(ConstantEntity):
    @classmethod
    def get_name(cls):
        return "CATEGORY"
    def get_description(self):
        return "Name of a configuration category"
    @classmethod
    def _request_values(cls, api):
        return api.get("config").keys()

class ObjType(MappedChoiceType):
    TYPES = OrderedDict((c.get_entity_short_name(), c) for c in (
        MasterVolume,       VolumeSnap,         VolumeClone,
        MasterFilesystem,   FilesystemSnap,     FilesystemClone,
        Pool,
        Host,
        Cluster,
        ))
    TYPE_NAMES = TYPES.keys()
    @classmethod
    def get_type(cls, name):
        return cls.TYPES[name.lower()]
    def _get_mapping(self):
        return self.TYPES

class EnvVar(Dynamic):
    def _get_dynamic(self, api, arg, value):
        env_var = os.environ.get(value)
        if env_var:
            return arg, env_var
        elif env_var is None:
            raise BadValue("ENV variable '%s' was not found." % (value))
        elif env_var == "":
            raise BadValue("ENV variable '%s' contains empty string" % (value))
        else:
            raise BadValue("ENV variable '%s' contains illegal characters" % (value))
    def get_description(self):
        return "Name of an existing environment variable"

### metrics related types

class MetricsSortByType(ChoiceType):
    TYPES = Munch(
        operations="ops",
        latence="latency",
        throughput="throughput",
        )
    def _get_values(self):
        return self.TYPES.values()
    @classmethod
    def get_name(cls):
        return "TYPE"
