from collections import OrderedDict

from infi.shell.commands import commandgroup, command, render
from infi.shell.types import (
    IPAddress, Hostname,
    )

from ..types import Name, EthPort
from .. import roles
from . import ConfigCommandsBase

@commandgroup(parent=ConfigCommandsBase, prefix="system")
class SystemConfigCommands(ConfigCommandsBase):

    def _format_ips_string(self, s):
        return (s or "").replace(",", ", ")

    @command(
        description="Display general system configuration",
        )
    @render(
        individually=True,
        )
    def query(self, _api):
        config = self._get_config(_api)
        result = OrderedDict((
            ("name", _api.get("system/name")),
            ("management_ip_address", "{} / {}".format(config.ip_config.ip, config.ip_config.netmask)),
            ("default_gateway", config.ip_config.gateway),
            ))
        node_ips = []
        for port in EthPort.iter_all(_api):
            if port.role and port.role.lower() == "management":
                node_ips.append((port.node_index, port.ip_v4_addr))
        result.update(("node_{}_ip_address".format(node_index), ip) for node_index, ip in sorted(node_ips))
        result.update((
            ("dns_servers", self._format_ips_string(config.mgmt["environment.dns_servers"])),
            ("ntp_servers", self._format_ips_string(config.mgmt["environment.ntp_servers"])),
            ))
        return result;

    @command(
        description="Set DNS server IPs",
        types=(IPAddress, IPAddress),
        )
    @roles.INFINIDAT.ADMIN
    def set_dns(self, _api, primary, secondary=None):
        servers = [primary.ip]
        if secondary:
            servers.append(secondary.ip)
        return self._configure_param(_api,
            category="mgmt",
            param="environment.dns_servers",
            value=",".join(servers),
            subject="DNS server" + ("s" if len(servers) > 1 else ""),
            value_name=", ".join(servers),
            )

    @command(
        description="Set system name",
        types=(Name,),
        )
    @roles.INFINIDAT.ADMIN
    def set_name(self, _api, name):
        current_name = _api.get("system/name")
        if current_name == name:
            return 'System name is already "{}"'.format(name)
        _api.put("system/name", name)
        return 'System renamed from "{}" to "{}"'.format(current_name, name)

@commandgroup(parent=SystemConfigCommands, prefix="time")
class SystemTimeCommands(ConfigCommandsBase):

    @command(
        description="Set NTP servers",
        types=(Hostname, Hostname),
        )
    @roles.INFINIDAT.ADMIN
    def set_ntp(self, _api, primary, secondary=None):
        servers = [primary]
        if secondary:
            servers.append(secondary)
        return self._configure_param(_api,
            category="mgmt",
            param="environment.ntp_servers",
            value=",".join(servers),
            subject="NTP server" + ("s" if len(servers) > 1 else ""),
            value_name=", ".join(servers),
            )
