from __future__ import print_function, absolute_import

import json
import requests
import httplib
import logging
import re
import os
import sys
from requests.packages import urllib3
from requests_toolbelt import MultipartEncoder
from munch import munchify, Munch
from colorama import Style, Fore
from itertools import chain
from functools import partial
from inspect import isroutine
from contextlib import contextmanager, nested as nested_contexts
from base64 import b64encode

from infi.shell import config
from infi.shell.shell import get_active_shell
from infi.shell.utils import Timer, normalize_sentence, NOTHING
from infi.shell.exceptions import CommandFailed, CommandAborted

from .. import debug
from ..utils import yesno, get_username_password, INPUT_INTERACTION_ERROR, Version
from ..version import VERSION as PRODUCT_VERSION
from .exceptions import (
    Unauthorized,                       # 401
    Forbidden,                          # 403
    NotFound,                           # 404
    ConfigurationConflict,              # 409
    InternalServerError,                # 500
    NotSupported,                       # 501
    ServiceTemporarilyUnavailable,      # 503
    ConnectionFailed,
    UnexpectedServerResponse,
    UnexpectedAPIResponse,
    ManagementException,
    BadPageStructure,
    )
from . import raw_mode

PRODUCT_NAME = "InfiniShell"

_REQUEST_TITLE = Style.BRIGHT + Fore.MAGENTA
_RESPONSE_TITLE = Style.BRIGHT + Fore.RED
_REQUEST_HEADERS = Fore.MAGENTA
_RESPONSE_HEADERS = Fore.RED
_REQUEST_DATA = Fore.GREEN + Style.BRIGHT
_TIME_SUMMARY = Style.BRIGHT + Fore.YELLOW
_ENDTITLE = Style.RESET_ALL
_EOL = Style.RESET_ALL

logger = logging.getLogger('api')

# Disable urllib3 warnings, which are triggered when a secure connection
# is established, but the certificate is not verified (by choice)
urllib3.disable_warnings()

class APIResult(object):
    def __init__(self, api, response):
        self.api = api
        self.response = response
    def is_success(self):
        return True
    def is_error(self):
        return False
    def get_error(self):
        return None
    def has_follow_up(self):
        return False
    def follow_up(self):
        return self
    def get_final_result(self):
        result = self
        while result.has_follow_up():
            result = result.follow_up()
        return result.get_result()
    def get_result(self):
        return self.response

class APISuccessResult(APIResult):
    def __init__(self, api, response):
        super(APISuccessResult, self).__init__(api, response)
        self.result = NOTHING
    def get_result(self):
        if self.result is not NOTHING:
            return self.result
        return self.response.data

class APIErrorResult(APIResult):
    def __init__(self, api, response):
        super(APIErrorResult, self).__init__(api, response)
        self.error_details = self.api._get_management_error_details(self.response)
    def is_success(self):
        return False
    def is_error(self):
        return True
    def get_error(self):
        return "{}: {}".format(self.error_details.code, self.error_details.formatted_message)
    def get_result(self):
        return self.api._handle_bad_response(self.response)

class APIApprovalRequired(APIResult):
    def __init__(self, api, response, req_kwargs):
        super(APIApprovalRequired, self).__init__(api, response)
        self.req_kwargs = dict(req_kwargs)
        self.req_kwargs.update(approved=True)
        self.approval_message, self.message_severity = self.api._get_approval_message(self.response)
    def is_success(self):
        return False
    def has_follow_up(self):
        return True
    def follow_up(self, auto_approve=False):
        if not auto_approve:
            # Confirm interactively and send again including confirmation header
            self.api._get_user_confirmation(self.approval_message,severity=self.message_severity)
        return self.api._send_request_get_result(**self.req_kwargs)
    def get_error(self):
        return "Aborted"
    def get_result(self):
        raise CommandAborted

class RemoteAPIBadCredentials(APIErrorResult):
    def get_error(self):
        return "Bad username or password for remote system"

class RestAPI(object):
    def __init__(self, baseuri, headers={}, encoding=None):
        self._baseuri = baseuri.rstrip("/") + "/"
        _headers = self._get_default_headers()
        _headers.update(headers)
        self._aggregator = None
        self.set_username(None)
        self.session = self._create_session(_headers)
        self._encoding = encoding
    def _create_session(self, headers):
        session = requests.session()
        session.headers.update(headers)
        return session
    def set_basic_auth(self, username, password):
        self.session.auth = (username, password)
        self.set_username(username)
    @contextmanager
    def change_basic_auth_context(self, username, password):
        old_auth = self.session.auth
        self.set_basic_auth(username, password)
        try:
            yield
        except:
            self.set_basic_auth(*old_auth)
            raise
    @contextmanager
    def no_basic_auth_context(self):
        auth, self.session.auth = self.session.auth, None
        try:
            yield
        finally:
            self.session.auth = auth
    def get_username(self):
        return self.username
    def set_username(self, username):
        self.username = username
    @classmethod
    def _get_default_headers(cls):
        return {}
    def _pre_send(self, data):
        return data if data is not NOTHING else None
    def _send(self, **kwargs):
        return_result_object = kwargs.pop("return_result_object", False)
        result = self._send_request_get_result(**kwargs)
        if not return_result_object:
            result = result.get_final_result()
        return result
    def _send_request_get_result(self, uri, method, data=NOTHING, headers=None, params=None, filenames=None, timeout=None):
        data = self._pre_send(data)
        full_uri = self._baseuri + uri
        response = self._send_request(method, full_uri, data, headers, params, filenames, timeout)
        self._post_receive(response)
        if self._is_bad_response(response):
            return APIErrorResult(self, response)
        return APIResult(self, response)
    def _send_request(self, method, uri, data, headers, params, filenames, timeout):
        assert (data, filenames).count(None) > 0, "data and filenames cannot be specified together"
        console_debugging = debug.is_on()
        logger_debugging = logger.isEnabledFor(logging.DEBUG)
        data, headers, file_contexts = self._get_multipart_params(data, headers, filenames)
        if console_debugging or logger_debugging:
            self._print_request_debug_info(**locals())
        try:
            timer = Timer()
            with nested_contexts(*file_contexts), timer:
                response_obj = self.session.request(
                    method,
                    uri,
                    params=params,
                    data=data,
                    headers=headers,
                    timeout=config.api_request_timeout_seconds if timeout is None else timeout,
                    verify=config.api_ssl_verify_certificate,
                    )
                response_obj.encoding = self._encoding
                response_data = response_obj.text
            response = Munch(
                code=response_obj.status_code,
                data=response_data,
                headers=response_obj.headers,
                response_object=response_obj,
                )
        except requests.Timeout as e:
            raise ConnectionFailed("Connection attempt timed out.")
        except requests.ConnectionError as e:
            error_message = unicode(e)
            try:
                if re.search(r'Name or service not known|nodename nor servname provided|getaddrinfo failed', error_message, re.I):
                    error_message = 'Could not resolve host name'
                else:
                    error_message = re.subn(r'^<.*? object at 0x.+?>: (Failed to establish a new connection: )?', '', e.args[0].reason.message)[0]
            except:
                pass
            raise ConnectionFailed(error_message)
        except IOError as e:
            raise ConnectionFailed(unicode(e))
        except Exception as e:
            raise ConnectionFailed(unicode(e))
        if console_debugging or logger_debugging:
            self._print_response_debug_info(**locals())
        if not console_debugging and raw_mode.is_on():
            shell = get_active_shell(safe=True)
            if shell and (shell.state.is_during_command() or shell.state.is_after_command()):
                self._print_debug(Fore.BLUE + uri + _EOL)
                self._print_debug(_REQUEST_DATA + response.data + _EOL)
        return response
    def _get_multipart_params(self, data, headers, filenames):
        file_contexts = []
        if filenames is not None:
            headers = dict(headers) if headers is not None else {}
            fields = {}
            for i, filename in enumerate(filenames, 1):
                f = open(filename, "rb")
                file_contexts.append(f)
                fields["file" + str(i)] = (os.path.basename(filename), f, "application/octet-stream")
            data = MultipartEncoder(fields=fields)
            headers["Content-Type"] = data.content_type
        return data, headers, file_contexts
    def _post_receive(self, response):
        return
    def _is_bad_response(self, response):
        return response.code >= 300
    _HTML_PATTERN = re.compile("<html>", re.IGNORECASE)
    def _handle_bad_response(self, response):
        code, data = response.code, response.data
        msg = None
        if data:
            if isinstance(data, dict):
                try:
                    msg = json.dumps(data)
                except BaseException:
                    msg = str(data)
            elif not isinstance(data, basestring):
                msg = str(data)
            elif self._HTML_PATTERN.match(data):
                msg = None
        self._raise_api_exception(code, msg)
    _exceptions = {
        401 : Unauthorized,
        403 : Forbidden,
        404 : NotFound,
        409 : ConfigurationConflict,
        500 : InternalServerError,
        501 : NotSupported,
        502 : ServiceTemporarilyUnavailable,  # Bad gateway means nothing to the user
        503 : ServiceTemporarilyUnavailable,
        }
    def _raise_api_exception(self, code, msg):
        AppropriateException = self._exceptions.get(code) or partial(UnexpectedServerResponse, code)
        raise AppropriateException(msg) if msg is not None else AppropriateException()
    def get(self, uri, params={}, **kwargs):
        return self._send(uri=uri, method="GET", params=params, **kwargs)
    def post(self, uri, data=NOTHING, **kwargs):
        return self._send(uri=uri, method="POST", data=data, **kwargs)
    def put(self, uri, data=NOTHING, **kwargs):
        return self._send(uri=uri, method="PUT", data=data, **kwargs)
    def delete(self, uri, data=NOTHING, **kwargs):
        return self._send(uri=uri, method="DELETE", data=data, **kwargs)
    def upload(self, uri, filenames, **kwargs):
        for filename in filenames:
            assert os.path.isfile(filename), "File does not exist: " + filename
        for arg in ("files", "headers"):
            if arg in kwargs:
                raise TypeError("Unsupported keyword argument for upload(): " + arg)
        return self._send(uri=uri, method="POST", filenames=filenames, **kwargs)
    def get_sub_resource(self, uri, subresource_class=None, parent_api=None, **extra_attrs):
        subresource_class = subresource_class or self.get_sub_resource_class()
        return subresource_class(parent_api or self, uri, **extra_attrs)
    @classmethod
    def get_sub_resource_class(cls):
        return APISubResource

    ## DEBUG PRINTS ##
    def _print_debug(self, msg, *args, **kwargs):
        if debug.is_on() or raw_mode.is_on():
            if args or kwargs:
                msg = msg.format(*args, **kwargs)
            if self._encoding:
                msg = msg.encode(self._encoding)
            print(msg, file=sys.stderr)
    def _format_headers(self, headers, color):
        max_length = max(len(k) for k in headers)
        template = "{}{:<" + str(max_length) + "} : {}{}{}"
        return "    " + "\n    ".join(template.format(color, k, Style.BRIGHT, v, _EOL) for k, v in headers.iteritems())
    def _format_data(self, data, color, with_title=True):
        if data:
            data = "<multipart file upload>" if isinstance(data, MultipartEncoder) else data
            return color + ("DATA: " if with_title else "") + _REQUEST_DATA + data + _EOL
        else:
            return color + "NO DATA" + _EOL
    __uri_colors_replacer = re.compile(r"([?&])")
    def _print_request_debug_info(_self, console_debugging, logger_debugging, uri, method, params, data, headers, **_):
        full_uri = uri
        if params:
            full_uri = "{}{}{}".format(uri, "&" if "?" in uri else "?",
                                       "&".join("{}={}".format(key, value)
                                       for key, value in params.iteritems()))
        full_uri = _self.__uri_colors_replacer.subn(Fore.CYAN + r"\1" + Fore.BLUE, full_uri)[0]
        request_info = _REQUEST_TITLE + u"### REQUEST: " + _ENDTITLE + \
                       Style.BRIGHT + Fore.CYAN + method + " " + Fore.BLUE + full_uri + _EOL
        all_headers = dict(chain(_self.session.headers.iteritems(), headers.iteritems()))
        request_headers_info = _self._format_headers(all_headers, _REQUEST_HEADERS)
        request_data_info = _self._format_data(data, _REQUEST_HEADERS)
        if logger_debugging:
            logger.info("API request:\n" + "\n".join((request_info, request_headers_info, request_data_info)))
        if console_debugging:
            _self._print_debug(request_info)
            _self._print_debug(request_headers_info)
            _self._print_debug(request_data_info)
    def _print_response_debug_info(_self, console_debugging, logger_debugging, response, timer, **_):
        response_info = _RESPONSE_TITLE + "### RESPONSE: " + _ENDTITLE + Style.BRIGHT + Fore.CYAN + \
                        "{} {}".format(response.code, httplib.responses[response.code]) + _EOL
        response_headers_info = _self._format_headers(response.headers, _RESPONSE_HEADERS)
        response_data_info = _self._format_data(response.data, _RESPONSE_HEADERS)
        response_summary = _TIME_SUMMARY + "Total time: {:.3f} seconds".format(timer.get_seconds()) + _EOL
        if logger_debugging:
            logger.info("API response:\n" + "\n".join((response_info, response_headers_info, response_data_info, response_summary)))
        if console_debugging:
            _self._print_debug(response_info)
            _self._print_debug(response_headers_info)
            _self._print_debug(response_data_info)
            _self._print_debug(response_summary)
            _self._print_debug("")

class APISubResource(object):
    _extra_attrs = ("_extra_attrs",)
    def __init__(self, api, base_uri, **extra_attrs):
        self._extra_attrs = set(chain(["_api", "_base_uri"], extra_attrs))
        self._api = api
        self._base_uri = base_uri.rstrip("/") + "/"
        for attr, value in extra_attrs.iteritems():
            setattr(self, attr, value)
    def get_sub_resource(self, uri, subresource_class=None, **extra_attrs):
        return self._api.get_sub_resource(uri, subresource_class, self, **extra_attrs)
    def __get_full_uri__(self, uri):
        return self._base_uri + uri.lstrip("/")
    def get(self, uri, *args, **kwargs):
        return self._api.get(self.__get_full_uri__(uri), *args, **kwargs)
    def post(self, uri, *args, **kwargs):
        return self._api.post(self.__get_full_uri__(uri), *args, **kwargs)
    def put(self, uri, *args, **kwargs):
        return self._api.put(self.__get_full_uri__(uri), *args, **kwargs)
    def delete(self, uri, *args, **kwargs):
        return self._api.delete(self.__get_full_uri__(uri), *args, **kwargs)
    def upload(self, uri, *args, **kwargs):
        return self._api.upload(self.__get_full_uri__(uri), *args, **kwargs)
    def __getattr__(self, attr):
        return getattr(self._api, attr)
    def __setattr__(self, attr, value):
        if attr not in self._extra_attrs:
            raise TypeError("Cannot set additional attributes in {}".format(self.__class__.__name__))
        return object.__setattr__(self, attr, value)

class InfiniboxAPI(RestAPI):
    _APPROVAL_HTTP_CODE = 403
    _HEADERS = Munch(
        CLIENT_ID = "X-Infinidat-Source",
        WARNINGS = "X-Infinidat-Warnings",
        REMOTE_AUTH = "X-Remote-Authorization",
        SYSTEM_VERSION = "X-Infinidat-Version",
        )
    _API_WARNING_SEPARATOR = ";"
    _PARAMS = Munch(
        APPROVED = "approved",
        )
    MGMT_ERROR_CODES = Munch(
        APPROVAL = "APPROVAL_REQUIRED",
        REMOTE_AUTH = "REMOTE_API_AUTHENTICATION_REQUIRED",
        )
    def __init__(self, baseuri, client_id, **kwargs):
        # Set client ID header
        kwargs.setdefault("headers", {})[self._HEADERS.CLIENT_ID] = client_id
        super(InfiniboxAPI, self).__init__(baseuri, **kwargs)
        self._api_warnings = set()
        self._remote_auth_context = None
        self.system_version = None
    @classmethod
    def get_sub_resource_class(cls):
        return InfiniboxAPISubResource
    @classmethod
    def _get_default_headers(cls):
        headers = super(InfiniboxAPI, cls)._get_default_headers()
        headers.update({"Content-Type": "application/json",
                        "Accept" : "application/json",
                        "Accept-Charset": "utf-8",
                        "User-Agent": "{}/{}".format(PRODUCT_NAME, PRODUCT_VERSION)})
        return headers
    @classmethod
    def _get_basic_authentication_value(cls, username, password):
        return "Basic " + b64encode("{}:{}".format(username, password))

    ## REQUEST AND RESPONSE LOGIC ##

    def _send(self, **kwargs):
        if self._aggregator and kwargs["method"].upper() != "GET":
            self._aggregator.add_request(**kwargs)
            return None
        return super(InfiniboxAPI, self)._send(**kwargs)
    def _send_request_get_result(self, **kwargs):
        """
        Additional keyword arguments:
        :param unwrap_result: Whether or not to unwrap the result from the envelope (default: True)
        :param approved: Whether or not the request is approved (default: config.auto_confirm_prompts)
        :param handle_remote_auth_error: Whether or not to ask the user for remote credentials (default: True)
        """
        unwrap_result = kwargs.pop("unwrap_result", True)
        is_approved = kwargs.pop("approved", config.auto_confirm_prompts)
        handle_remote_auth_error = kwargs.pop("handle_remote_auth_error", True)
        if is_approved:
            # Set approved header
            params = kwargs.get("params") or {}
            params[self._PARAMS.APPROVED] = "true"
            kwargs["params"] = params
        headers = kwargs.setdefault("headers", {})
        self._apply_remote_credentials(headers)
        # Send request
        apiresult = super(InfiniboxAPI, self)._send_request_get_result(**kwargs)
        self._set_system_version(apiresult.response)
        # Handle remote authentication and resend if needed
        if handle_remote_auth_error and apiresult.is_error() and self._remote_auth_context and self._is_remote_auth_error(apiresult.response):
            self._set_and_apply_remote_credentials(headers)
            apiresult = super(InfiniboxAPI, self)._send_request_get_result(**kwargs)
            if apiresult.is_error() and self._is_remote_auth_error(apiresult.response):
                return RemoteAPIBadCredentials(apiresult.api, apiresult.response)
        # Handle approval response
        if apiresult.is_error() and not is_approved and self._is_approval_required(apiresult.response):
            return APIApprovalRequired(self, apiresult.response, kwargs)
        # Handle success
        if apiresult.is_success():
            apiresult = APISuccessResult(self, apiresult.response)
            if unwrap_result:
                apiresult.result = self._unwrap_result(apiresult.response.data)
        return apiresult

    def _set_system_version(self, response):
        if not self.system_version:
            version = response.headers.get(self._HEADERS.SYSTEM_VERSION)
            self.system_version = Version(version)

    def _get_management_error_details(self, response):
        e = getattr(response, "_management_error_details", None)
        if e is None:
            e = response._management_error_details = ManagementErrorDetails(response)
        return e

    def _get_approval_message(self, response):
        message = self._get_management_error_details(response)
        return message.formatted_message, message.severity
    def _is_approval_required(self, response):
        return response.code == self._APPROVAL_HTTP_CODE and self._get_management_error_details(response).code.startswith(self.MGMT_ERROR_CODES.APPROVAL)
    def _get_user_confirmation(self, msg, abort=True, severity=""):
        msg = msg or 'This action requires your confirmation.'
        if severity in  ["HIGH","ERROR"]:
            get_active_shell().error(msg)
        else:
            get_active_shell().warning(msg)
        if not yesno("Are you sure?"):
            if abort:
                raise CommandAborted
            return False
        return True

    @contextmanager
    def remote_auth_context(self, system_name=None, remote_api=None):
        assert any((system_name, remote_api))
        old_value, self._remote_auth_context = self._remote_auth_context, Munch(
            username=None,
            password=None,
            credentials_set=False,
            system_name=system_name,
            remote_api=remote_api,
            )
        try:
            yield self._remote_auth_context
        finally:
            self._remote_auth_context = old_value
    def _apply_remote_credentials(self, headers):
        ctx = self._remote_auth_context
        if ctx and (ctx.remote_api or ctx.username):
            value = ctx.remote_api.get_basic_auth_value() if ctx.remote_api \
                    else self._get_basic_authentication_value(ctx.username, ctx.password or "")
            if value is not None:
                headers[self._HEADERS.REMOTE_AUTH] = value
    def _set_and_apply_remote_credentials(self, headers):
        ctx = self._remote_auth_context
        assert ctx, "Not inside a remote authentication context"
        ctx.username, ctx.password = get_username_password(
            prompt="Username for {}".format(ctx.system_name),
            interaction_error=INPUT_INTERACTION_ERROR + ", or invoke link.authenticate prior to this command")
        ctx.credentials_set = True
        if ctx.remote_api:
            ctx.remote_api.set_basic_auth(ctx.username, ctx.password)
        self._apply_remote_credentials(headers)
    def _is_remote_auth_error(self, response):
        return self._get_management_error_details(response).code == self.MGMT_ERROR_CODES.REMOTE_AUTH

    def _unwrap_result(self, data):
        if not isinstance(data, dict) or "result" not in data:
            raise UnexpectedAPIResponse("Expected wrapped response, but got this: {}".format(data))
        return data["result"]
    def _pre_send(self, data):
        return json.dumps(data, indent=2) if data is not NOTHING else None
    def _post_receive(self, response):
        super(InfiniboxAPI, self)._post_receive(response)
        for attr in (a for a in dir(self) if a.startswith("_post_receive__")):
            method = getattr(self, attr)
            if isroutine(method):
                method(response)
    def _post_receive__body(self, response):
        if response.headers.get("content-type") == "application/json":
            try:
                json_data = json.loads(response.data)
            except ValueError:
                return
            response.data = munchify(json_data)
    def _post_receive__api_warnings(self, response):
        if not config.api_warnings:
            return
        shell = get_active_shell(safe=True)
        # No point in doing this unless shell is active, and only during
        # command execution, i.e not during auto-completion and such
        if not shell or not shell.state.is_during_command():
            return
        api_warnings = response.headers.get(self._HEADERS.WARNINGS)
        if api_warnings:
            for warning in api_warnings.split(self._API_WARNING_SEPARATOR):
                warning = warning.strip()
                if warning and warning not in self._api_warnings:
                    self._api_warnings.add(warning)
                    shell.error("API warning: " + warning, colors=[Style.BRIGHT, Fore.YELLOW])
    def _handle_bad_response(self, response):
        error = response.data
        if isinstance(error, dict) and "error" in error:
            error = self._get_management_error_details(response)
            raise ManagementException(response.code, error)
        super(InfiniboxAPI, self)._handle_bad_response(response)

    ## REQUEST AGGREGATION ##

    @contextmanager
    def _request_aggregation_context(self, aggregator):
        assert not self._aggregator, "Already inside aggregation context"
        self._aggregator = aggregator
        try:
            yield self._aggregator
        finally:
            self._aggregator = None
    def create_request_aggregator(self, *args, **kwargs):
        return RequestAggregator(self, *args, **kwargs)
    @contextmanager
    def aggregation(self, *args, **kwargs):
        aggregator = self.create_request_aggregator(*args, **kwargs)
        with aggregator:
            yield aggregator
        aggregator.run()
    def is_aggregating(self):
        return self._aggregator is not None

    ## API FEATURE METHODS ##

    def ping(self, base_uri=None):
        return self.get("rest/system/ping", headers={"Content-Type": "text/plain", "Accept": "*"}, unwrap_result=False)
    def paged_query(self, uri, page_size=None, params=None, headers=None, hook=None):
        return PagedQuery(self, uri, page_size or config.api_default_page_size, params=params, headers=headers, hook=hook)
    def create_filter(self, **kwargs):
        """
        Create a query filter object. Additional keyword arguments are treated as
        equality filters, and the returned filter is set accordingly.

        :param: kwargs keyword arguments for constructor
        :returns: :py:class:`QueryFilter`
        """
        return QueryFilter(**kwargs)
    def create_sort(self):
        """
        Create a query sort object.

        :returns: :py:class:`QuerySort`
        """
        return QuerySort()
    def create_pluck(self, *fields):
        """
        Create a query pluck object.

        :returns: :py:class:`QueryPluck`
        """
        return QueryPluck(*fields)

class ManagementErrorDetails(object):
    _stacktrace_line_pattern = re.compile("^([^\tat]|\tat com.infinidat)")
    def __init__(self, response):
        self._parse_response(response)
        self.formatted_message = normalize_sentence(self.message or self.code, end_with=":" if self.reasons else ".")
        if self.reasons:
            self.formatted_message = "{}{}{}".format(
                self.formatted_message,
                "\n  " if len(self.reasons) > 1 else " ",
                "\n  ".join(normalize_sentence(r.message) for r in self.reasons))
        if debug.is_on() and response.code >= 500 and self.stack:
            self.formatted_message = "{}\n\nServer stack trace:\n{}".format(self.formatted_message, self._format_stack())
    def _parse_response(self, response):
        if isinstance(response.data, dict) and isinstance(response.data.get("error"), dict):
            e = response.data["error"]
        else:
            e = {}
        for member, field, expected_type, default in (
            ("code", "code", basestring, "<MISSING CODE>"),
            ("message", "message", basestring, None),
            ("reasons", "reasons", (list, tuple, set, dict), ()),
            ("stack", "stack", basestring, None),
            ("severity", "severity", basestring, None),
            ("is_remote", "is_remote", bool, False),
            ):
            value = e.get(field)
            if not isinstance(value, expected_type):
                value = default
            setattr(self, member, value)
    def _format_stack(self):
        lines = []
        truncating = False
        for line in self.stack.splitlines(True):
            if self._stacktrace_line_pattern.match(line):
                truncating = False
                lines.append(line)
            elif not truncating:
                truncating = True
                lines.append("\tat ...\n")
        return "".join(lines)

class ParamFactory(object):
    """ Base class for query parameter factories """
    def to_params(self):
        raise NotImplementedError

class QueryFilter(ParamFactory):
    """ Query parameter factory for filtering. Instances are immutable, so
    methods such as eq(), gt(), between(), and so on return a new instance
    each time. """
    LIST_SEPARATOR = ","
    SIMPLE_OPS = {op : op for op in "eq ne lt le gt ge like".split()}
    RANGE_OPS = dict(
        member_of="in",
        not_member_of="out",
        between="between",
        )
    class __metaclass__(type):
        def __new__(mcls, name, bases, attrs):
            def _add_func(_name, f, doc):
                f.__name__ = _name
                f.__doc__ = doc
                attrs[_name] = f
            def _create_simple_op_func(func_name, operator):
                return lambda self, field, value: self._add(field, operator, value)
            def _create_range_op_func(func_name, operator):
                return lambda self, field, *values: self._add(field, operator, *values)
            for func_name, operator in attrs["SIMPLE_OPS"].iteritems():
                _add_func(
                    func_name,
                    _create_simple_op_func(func_name, operator),
                    "{} operator (returns new filter)".format(operator.upper()),
                    )
            for func_name, operator in attrs["RANGE_OPS"].iteritems():
                _add_func(
                    func_name,
                    _create_range_op_func(func_name, operator),
                    "{} operator (returns new filter)".format(operator.upper()),
                    )
            return type.__new__(mcls, name, bases, attrs)
    def __init__(self, **kwargs):
        """ Keyword arguments are translated to equality filters """
        self._filter = {}
        for field, value in kwargs.iteritems():
            self._add_mutation(field, self.SIMPLE_OPS["eq"], value)
    def to_params(self):
        return dict(self._filter)
    def _add(self, field, operator, *values):
        assert field not in self._filter, "Attempted to filter the same field twice: " + field
        new_filter = QueryFilter()
        new_filter._filter = dict(self._filter)
        new_filter._add_mutation(field, operator, *values)
        return new_filter
    def _add_mutation(self, field, operator, *values):
        self._filter[field] = "{}:{}".format(operator, self.LIST_SEPARATOR.join(str(v) for v in values))

class QuerySort(ParamFactory):
    """ Query parameter factory for sorting. Instances are immutable, so asc()
    and desc() return new instances. """
    LIST_SEPARATOR = ","
    ASC = ""  # Empty operator
    DESC = "-"
    def __init__(self):
        self._sort_string = ""
        self._sort_fields = ()
    def sort(self, field, is_ascending):
        new_sort = type(self)()
        new_sort._sort_string = (self.ASC if is_ascending else self.DESC) + field
        if self._sort_string:
            new_sort._sort_string = "{}{}{}".format(self._sort_string, self.LIST_SEPARATOR, new_sort._sort_string)
        new_sort._sort_fields = tuple(chain(self._sort_fields, ((field, is_ascending),)))
        return new_sort
    def asc(self, field):
        return self.sort(field, True)
    def desc(self, field):
        return self.sort(field, False)
    def to_params(self):
        return dict(sort=self._sort_string or None)
    def sort_entities(self, entities):
        assert isinstance(entities, list), "Entities argument expected to be a list, but is a {!r}".format(type(entities))
        for field, is_ascending in reversed(self._sort_fields):
            entities.sort(key=lambda entity: entity[field], reverse=not is_ascending)
    def sorted_entities(self, entities):
        entities = list(entities)
        self.sort_entities(entities)
        return entities

class QueryPluck(ParamFactory):
    """ Query parameter factory for field-plucking. Instances are immutable, so
    pluck() returns a new instance. """
    LIST_SEPARATOR = ","
    def __init__(self, *fields):
        self._fields = fields
    def pluck(self, *fields):
        return type(self)(*(self._fields + fields))
    def to_params(self):
        return dict(fields=self.LIST_SEPARATOR.join(self._fields) or None)

class InfiniboxAPISubResource(APISubResource):
    def paged_query(self, uri, *args, **kwargs):
        return self._api.paged_query(self.__get_full_uri__(uri), *args, **kwargs)

class PagedQuery(object):
    _PAGE_FIELDS = ("result", "metadata")
    def _assert_params(self, params):
        for param in ("page", "page_size"):
            assert param not in params, "'{}' is a reserved parameter, and cannot be specified in 'params'".format(param)
        return True
    def __init__(self, api, url, page_size=None, params=None, headers=None, hook=None):
        params = params or {}
        assert self._assert_params(params)
        self.api = api
        self.url = url
        self._objects_hook = hook
        self._params = Munch(params)
        self._headers = headers or {}
        if page_size is not None:
            self._params.page_size = page_size
        self._first_page = self._get_page(1)
        self.page_size = self._first_page.metadata.page_size
        self.total_pages = self._first_page.metadata.pages_total
        self.total_objects = self._first_page.metadata.number_of_objects
        # Update params for next pages
        self._params.page_size = self.page_size
    def __iter__(self):
        return self.iter_all_objects()
    def iter_pages(self):
        for number in xrange(1, self.total_pages+1):
            yield self.get_page(number)
    def _get_page(self, number):
        self._params.page = number
        page_data = self.api.get(self.url, params=self._params, headers=self._headers, unwrap_result=False)
        self._assert_page_data(page_data)
        if self._objects_hook:
            page_data.result = self._objects_hook(page_data.result)
        return page_data
    def _assert_page_data(self, page_data):
        if not isinstance(page_data, dict) or not all(f in page_data for f in self._PAGE_FIELDS):
            raise BadPageStructure(page_data)
    def get_page(self, number):
        assert number >= 1, "Page numbers are 1-based"
        page_data = self._first_page if number == 1 else self._get_page(number)
        return Page(self, number, page_data.result)
    def get_all_objects(self):
        return list(self.iter_all_objects())
    def iter_all_objects(self):
        for page in self.iter_pages():
            for object in page:
                yield object

class Page(object):
    def __init__(self, query, number, objects):
        self.query = query
        self.number = number
        self.objects = objects
    def __iter__(self):
        return iter(self.objects)

class AggregatedTask(object):
    def __init__(self, name, success_msg):
        self.name = name
        self.success_msg = success_msg
        self.requests = []
        self.last_result = None
        self.state = None
    def get_success_message(self):
        assert self.last_result and self.last_result.is_success()
        if callable(self.success_msg):
            return self.success_msg(self.last_result.get_result())
        return self.success_msg

class RequestAggregator(object):
    SUCCESS, FAILURE, APPROVAL = xrange(3)
    def __init__(self, api, total_success_msg=None):
        self._ctx = api._request_aggregation_context(self)
        self._api = api
        self._tasks = []
        self._current_task = None
        self.results = []
        self._used = False
        self._total_success_msg = total_success_msg
    def __enter__(self):
        self._ctx.__enter__()
    def __exit__(self, *args, **kwargs):
        self._ctx.__exit__(*args, **kwargs)
    def is_empty(self):
        return not self._tasks or len(self._tasks) == 1 and not self._current_task.requests and self._current_task.state is None
    def _assert_not_used(self):
        assert not self._used, "Reusing an aggregator is prohibited - create a new one"
    def _assert_current_task(self):
        assert self._current_task, "No current task: invoke next_task() before invoking API calls"
    def add_request(self, *args, **kwargs):
        self._assert_not_used()
        self._assert_current_task()
        request = partial(self._api._send_request_get_result, *args, **kwargs)
        self._current_task.requests.append(request)
    def _remove_current_task_if_empty(self):
        if self._current_task and not self._current_task.requests:
            self._tasks.pop()
    def next_task(self, task_name, success_msg):
        """
        :param task_name: Task name, used as a prefix for the success message
        :param success_msg: Success message (string or ``func(result)``)
        """
        self._assert_not_used()
        self._remove_current_task_if_empty()
        self._current_task = AggregatedTask(task_name, success_msg)
        self._tasks.append(self._current_task)
    def run(self):
        self._assert_not_used()
        self._assert_current_task()
        self._used = True
        self._remove_current_task_if_empty()
        tasks_remaining = True
        successes, failures = [], []
        while tasks_remaining:
            tasks_remaining = False
            pending_approval = []
            for task in self._tasks:
                if task.state in (self.SUCCESS, self.FAILURE):
                    continue
                while task.requests:
                    # Get next request for this task
                    request = task.requests.pop(0)
                    # Perform request
                    result = task.last_result = request()
                    # If result requires follow-up, get subsequent results, unless the follow-up entails approval
                    while result.has_follow_up() and not isinstance(result, APIApprovalRequired):
                        result = result.follow_up()
                    task.last_result = result
                    # If result is a failure, mark task as "failed" and don't send subsequent requests
                    if result.is_error():
                        task.state = self.FAILURE
                        failures.append(task)
                        break
                    # If approval is required, mark task as "approval", and don't send subsequent requests
                    if isinstance(result, APIApprovalRequired):
                        task.state = self.APPROVAL
                        pending_approval.append(task)
                        break
                    # Result is a success - carry on
                    assert result.is_success(), "Unhandled result type: {}".format(result)
                # If there are no more requests for this task, and everything went OK, mark task as "success"
                if not task.requests and not task.state:
                    task.state = self.SUCCESS
                    successes.append(task)
                # If task state is not a terminal state, then another iteration is needed
                if task.state not in (self.SUCCESS, self.FAILURE):
                    tasks_remaining = True
            # Handle tasks pending approval
            if pending_approval:
                # Get user confirmation to approve all tasks
                if not self._api._get_user_confirmation(
                    "\n".join(self._iter_task_messages(
                        pending_approval,
                        lambda task: task.last_result.approval_message)),
                    abort=False,
                    severity=task.last_result.response._management_error_details.severity,
                    ):
                    # User declined, so consider all unapproved tasks as failures
                    for task in pending_approval:
                        task.state = self.FAILURE
                        failures.append(task)
                    break
                # Push follow-up request to each approved task (at this point they're all approved)
                for task in pending_approval:
                    task.state = None
                    task.requests.insert(0, partial(task.last_result.follow_up, auto_approve=True))
        # If only one task was aggregated, and it failed
        if len(self._tasks) == 1:
            if failures:
                failures[0].last_result.get_result()  # This raises an exception
            # Return the total success message, or the single success message of the one task
            return self._get_total_success_message(successes) or successes[0].get_success_message()
        # If there are no failures, return the total success message, or a successful batch result
        if not failures:
            return self._get_total_success_message(successes) or "\n".join(self._iter_success_messages(successes))
        # Result consists of both successes and failures, so return a nested list of successes and failures
        listing = NestedListing()
        listing.append("")
        if successes:
            with listing.nested():
                listing.append("Successful actions:")
                with listing.nested():
                    listing.extend(self._iter_success_messages(successes))
        with listing.nested():
            listing.append("Failed actions:")
            with listing.nested():
                failure_msgs = self._iter_task_messages(failures, lambda task: task.last_result.get_error())
                listing.extend(failure_msgs)
        raise CommandFailed(listing.render())
    def _get_total_success_message(self, successes):
        if not self._total_success_msg:
            return None
        if callable(self._total_success_msg):
            return self._total_success_msg(task.last_result.get_result() for task in successes)
        return self._total_success_msg
    def _iter_task_messages(self, tasks, msg_func):
        return ("{}: {}".format(task.name, msg_func(task)) for task in tasks)
    def _iter_success_messages(self, successes):
        return self._iter_task_messages(successes, lambda task: task.get_success_message())

class NestedListing(object):
    def __init__(self):
        self.items = []
        self.level = 0
    def append(self, s):
        self.items.append((self.level, s))
    def extend(self, strings):
        self.items.extend((self.level, s) for s in strings)
    @contextmanager
    def nested(self):
        self.level += 1
        try:
            yield
        finally:
            self.level -= 1
    def render(self, indent=2):
        from StringIO import StringIO
        stream = StringIO()
        self.render_to(stream, indent=indent)
        stream.seek(0)
        return stream.read()
    def render_to(self, stream, indent=2):
        indentations = {}
        for i, (level, item) in enumerate(self.items):
            if i != 0:
                stream.write("\n")
            indentation = indentations.get(level)
            if indentation is None:
                indentation = indentations[level] = " " * indent * level
            for j, line in enumerate(item.split("\n")):
                if j != 0:
                    stream.write("\n")
                stream.write(indentation)
                stream.write(line)
