from contextlib import contextmanager
from numbers import Number

from ..utils import INPUT_INTERACTION_ERROR
from .api import InfiniboxAPISubResource
from .system import System

REPLICATION_INPUT_INTERACTION_ERROR = INPUT_INTERACTION_ERROR + ", or invoke link.authenticate"

_remote_systems_cache = dict()

def clear_remote_system_cache():
    _remote_systems_cache.clear()

def get_remote_system(_api, link_id_or_link_or_replica):
    """ Get a remote system instance, that uses a remote API instance. A remote
    API instance features all API methods, but works against a remote system's
    API via delegation (/api/rest/remote/...).

    This API object offers a context:
    ... with get_remote_system(api, link_id_or_link_or_replica).api as remote_api:
    ...    remote_api.get(...)
    ...    api.post(...)

    During the context of a remote API object, remote authentication details are
    applied also to its parent (local) API instance.
    """
    link_id, link = None, None
    if isinstance(link_id_or_link_or_replica, Number):
        # This is a link ID
        link_id = link_id_or_link_or_replica
    elif "link_id" in link_id_or_link_or_replica:
        # This is a replica entity
        link_id = link_id_or_link_or_replica.link_id
    else:
        # This is a link entity
        link_id, link = link_id_or_link_or_replica.id, link_id_or_link_or_replica
    assert link_id is not None
    system = _remote_systems_cache.get(link_id)
    if not system:
        remote_api = _api.get_sub_resource(
            "remote/{}/api/rest".format(link_id),
            subresource_class=RemoteInfiniboxAPISubResource,
            _link_id=link_id,
            _link=link,
            )
        system = _remote_systems_cache[link_id] = System(remote_api)
    return system

class RemoteInfiniboxAPISubResource(InfiniboxAPISubResource):

    def __init__(self, *args, **extra_attrs):
        extra_attrs.update(
            _remote_auth_context=None,
            _remote_auth_value=None,
            )
        super(RemoteInfiniboxAPISubResource, self).__init__(*args, **extra_attrs)

    def get_link(self):
        if not self._link:
            from .commands.types import ReplicationLinkBySystemName  # Avoid cyclic dependencies
            self._link = ReplicationLinkBySystemName.get_by_id(self._api, self._link_id)
        return self._link

    def get_remote_system_name(self):
        return self.get_link().remote_system_name

    @contextmanager
    def no_basic_auth_context(self):
        # Remote auth should never be disabled
        yield

    def set_basic_auth(self, username, password):
        self.set_basic_auth_value(self._get_basic_authentication_value(username, password))

    def set_basic_auth_value(self, value):
        self._remote_auth_value = value

    def get_basic_auth_value(self):
        return self._remote_auth_value

    def __enter__(self):
        self._remote_auth_context = self._api.remote_auth_context(self.get_remote_system_name(), self)
        self._remote_auth_context.__enter__()
        return self

    def __exit__(self, *args, **kwargs):
        self._remote_auth_context.__exit__(*args, **kwargs)
