import re
import datetime
import capacity
import math
from numbers import Number

from ...utils import PossiblyNone

__all__ = ['ValueRenderer']

class ValueRenderer(object):
    RENDERERS = None # Value is set at module bottom
    def __init__(self, empty_value="-"):
        self._empty_value = empty_value
    def render(self, field, value, **properties):
        if isinstance(value, PossiblyNone) and not value._is_none:
            value = value._value
        if value is None or value == "" or (hasattr(value, "__len__") and not value):
            return self._empty_value
        try:
            for renderer in self.RENDERERS:
                if renderer.is_suitable(field, value):
                    return unicode(renderer.render(value, self._empty_value, **properties))
        except Exception as e:
            return u"[{}]".format(e).upper()
        return unicode(value)
    def is_left_aligned(self, field):
        """
        :returns: ``True`` for left alignment, ``False`` for right alignment,
                  ``None`` if undefined.
        """
        for renderer in self.RENDERERS:
            if renderer.is_field_explicitly_suitable(field):
                return renderer.LEFT_ALIGNED
        return None

# === ABSTRACT VALUE RENDERERS ===

class ValueTypeRenderer(object):
    LEFT_ALIGNED = True
    def is_field_explicitly_suitable(self, field):
        return False
    def is_suitable(self, field, value):
        raise NotImplementedError
    def render(self, obj, empty_value, **properties):
        raise NotImplementedError

class PatternSearchValueTypeRenderer(ValueTypeRenderer):
    class __metaclass__(type):
        def __new__(mcls, name, bases, attrs):
            cls = type.__new__(mcls, name, bases, attrs)
            if isinstance(cls.PATTERN, basestring):
                cls.PATTERN = re.compile(cls.PATTERN, re.IGNORECASE)
            return cls
    PATTERN = None
    VALUE_TYPE = None
    def is_field_explicitly_suitable(self, field):
        return self._is_matching(field)
    def is_suitable(self, field, value):
        if self.VALUE_TYPE and not isinstance(value, self.VALUE_TYPE):
            return False
        return self.is_field_explicitly_suitable(field)
    def _is_matching(self, field):
        return self.PATTERN.search(field) is not None

# === SOLID VALUE RENDERERS ===

class CompoundObjectRenderer(ValueTypeRenderer):
    def is_suitable(self, field, value):
        return isinstance(value, (dict, list, set, tuple))
    def render(self, obj, empty_value, **properties):
        return "<compound>"

class RateLimitRenderer(PatternSearchValueTypeRenderer):
    PATTERN = "RATE_LIMIT"
    VALUE_TYPE = Number
    LEFT_ALIGNED = False
    def render(self, value, empty_value, **properties):
        if value in [0,]:
            return empty_value
        else:
            return "{}Mbps".format(value)

class StorageCapacityRenderer(PatternSearchValueTypeRenderer):
    PATTERN = "(SIZE|CAPACITY|SPACE|DISK_USAGE|MAX_EXTEND|ALLOCATED|USED|CONSUMED)"
    VALUE_TYPE = Number
    LEFT_ALIGNED = False
    _capacities = ((float(capacity.TB / capacity.byte), "TB"),
                   (float(capacity.GB / capacity.byte), "GB"),
                   (float(capacity.MB / capacity.byte), "MB"),
                   (float(capacity.KB / capacity.byte), "KB"))
    _binary_capacities = ((float(capacity.TiB / capacity.byte), "TiB"),
                          (float(capacity.GiB / capacity.byte), "GiB"),
                          (float(capacity.MiB / capacity.byte), "MiB"),
                          (float(capacity.KiB / capacity.byte), "KiB"))
    def render(self, bytes, empty_value, **properties):
        if properties.get("capacity_unit"):
            capacity_unit, unit = properties.get("capacity_unit")
            if unit == "B":
                return bytes
            unit_bytes = float(capacity_unit / capacity.byte)
            computed_value = bytes / unit_bytes
            if unit == "blocks":
                computed_value = int(math.ceil(computed_value))  # Round up, since blocks cannot be fractioned
                formatter = "{} {}"
            else:
                formatter = "{{:.{}f}} {{}}".format(3 if unit[0] in ("T") else 2)
            return formatter.format(computed_value, unit)
        capacities = self._binary_capacities if properties.get("binary_capacity") else self._capacities
        for i, (unit_bytes, unit) in enumerate(capacities):
            if bytes >= unit_bytes or round(bytes / (unit_bytes / 1000), 2) >= 1000:
                # Bytes should be rounded manually
                return "{} {}".format(self._truncate_float(bytes / unit_bytes, 2), unit)
        return bytes
    def _truncate_float(self, f, num_decimal_digits):
        return ('%.*f' % (num_decimal_digits + 1, f))[:-1]

class DataBitRateRenderer(PatternSearchValueTypeRenderer):
    PATTERN = "(CONNECTION_SPEED|MAX_SPEED)"
    VALUE_TYPE = Number
    LEFT_ALIGNED = False
    _rates  = ((10**12, "Tbps"),
               (10**9,  "Gbps"),
               (10**6,  "Mbps"),
               (10**3,  "Kbps"),
               (1,       "bps"))
    def render(self, value, empty_value, **properties):
        if value == 0:
            return "-"
        if value:
            for rate, unit in self._rates:
                if value >= rate:
                    return "{} {}".format(value / rate, unit)

class DataByteRateRenderer(DataBitRateRenderer):
    PATTERN = "(THROUGHPUT)"
    _rates  = ((10**12, "TB/sec"),
               (10**9,  "GB/sec"),
               (10**6,  "MB/sec"),
               (10**3,  "KB/sec"),
               (1,       "B/sec"))

class StorageCapacityTerabytesRenderer(StorageCapacityRenderer):
    def render(self, bytes, empty_value, **properties):
        bytes = bytes or 0
        return "{:.3f} TB".format(bytes/(1000.0**4))

class MillisecondsDeltaRenderer(PatternSearchValueTypeRenderer):
    PATTERN = "(KEEP_ALIVE|TIMEOUT|SYNC_INTERVAL|LAST_SYNC_DURATION|RPO_VALUE)"
    VALUE_TYPE = Number
    def render(self, timestamp, empty_value, **properties):
        timestamp = timestamp/1000.0
        m, s = divmod(timestamp, 60)
        h, m = divmod(m, 60)
        return "%d:%02d:%02d" % (h, m, s)


class UpTimeRenderer(MillisecondsDeltaRenderer):
    PATTERN = 'UPTIME'

    @staticmethod
    def _plural(name, value):
        return '%d %s%s' % (value, name, '' if value == 1 else 's')

    def render(self, msecs, empty_value, **properties):
        seconds = msecs / 1000
        minutes = seconds / 60
        hours = minutes / 60
        days = hours / 24
        months = days / 30
        years = months / 12

        seconds %= 60
        minutes %= 60
        hours %= 24

        parts = []
        if years:
            parts.append(UpTimeRenderer._plural('year', years))

        if months or parts:
            parts.append(UpTimeRenderer._plural('month', months))

        if days or parts:
            parts.append(UpTimeRenderer._plural('day', days))

        parts.append(UpTimeRenderer._plural('hour', hours))
        parts.append(UpTimeRenderer._plural('minute', minutes))

        return ' '.join(parts)


class MillisecondsTimestampRenderer(PatternSearchValueTypeRenderer):
    PATTERN = "(TIME|DATE|(CREATED|MODIFIED)_AT|LAST_SYNCHRONIZED|RESTORE_POINT|(INSTALLED|UPDATED)_ON)"
    VALUE_TYPE = Number
    def render(self, timestamp, empty_value, **properties):
        if isinstance(timestamp, basestring):
            timestamp = long(timestamp)
        return datetime.datetime.fromtimestamp(timestamp/1000.0).strftime("%Y-%m-%d %H:%M:%S")

class SimpleArrayRenderer(ValueTypeRenderer):
    def is_suitable(self, field, value):
        if not isinstance(value, (list, set, tuple)):
            return False
        for v in value:
            if not isinstance(v, (Number, basestring)):
                return False
        return True
    def render(self, obj, empty_value, **properties):
        return u", ".join(unicode(v) for v in obj)

class BooleanRenderer(ValueTypeRenderer):
    def is_suitable(self, field, value):
        return value is True or value is False
    def render(self, obj, empty_value, **properties):
        return u"yes" if obj is True else u"no"

# Order is important; the first suitable renderer is used
ValueRenderer.RENDERERS = (
    BooleanRenderer(),
    StorageCapacityRenderer(),
    UpTimeRenderer(),
    MillisecondsDeltaRenderer(),
    MillisecondsTimestampRenderer(),
    SimpleArrayRenderer(),
    CompoundObjectRenderer(),
    RateLimitRenderer(),
    DataByteRateRenderer(),
    DataBitRateRenderer(),
    )
